"""
dgcv: Differential Geometry with Complex Variables

This module defines defines functions for interacting with the Variable Management Framework (VMF), 
which is dgcv's system for managing object creation and labeling. It additionaly contains functions
for internal use by the VMF. The intended public functions include the following

Functions for listing and clearing objects in the VMF:
    - listVar(): Lists the "parent names" of objects currently tracked by the dgcv VMF.
    - clearVar(): Clears the variables from the dgcv registry and deletes them from caller's globals().

Functions for summarizing the state of the VMF:
    - vmf_summary(): Takes a snapshot of the current dgcv VMF and reports a summary in a
    Pandas table.

Author: David Sykes (https://github.com/YikesItsSykes)

Dependencies:
    - sympy
    - pandas

License:
    MIT License
"""

import re
import warnings

import pandas as pd
import sympy as sp
from IPython.display import display
from sympy import I

from ._config import (
    _cached_caller_globals,
    get_dgcv_settings_registry,
    get_variable_registry,
    greek_letters,
)
from .combinatorics import carProd_with_weights_without_R, permSign
from .conversions import _allToSym
from .styles import get_style


def _coeff_dict_formatter(
    varSpace,coeff_dict,valence,total_degree,_varSpace_type,data_shape
):
    """
    Helper function to populate conversion dicts for tensor field classes
    """
    variable_registry = get_variable_registry()
    CVS = variable_registry["complex_variable_systems"]

    exhaust1 = list(varSpace)
    populate = {
        "compCoeffDataDict": dict(),
        "realCoeffDataDict": dict(),
        "holVarDict": dict(),
        "antiholVarDict": dict(),
        "realVarDict": dict(),
        "imVarDict": dict(),
        "preProcessMinDataToHol": dict(),
        "preProcessMinDataToReal": dict(),
    }
    if _varSpace_type == "real":
        for var in varSpace:
            varStr = str(var)
            if var in exhaust1:
                for parent in CVS.values():
                    if varStr in parent["variable_relatives"]:
                        cousin = (
                            set(
                                parent["variable_relatives"][varStr][
                                    "complex_family"
                                ][2:]
                            )
                            - {var}
                        ).pop()
                        if cousin in exhaust1:
                            exhaust1.remove(cousin)
                        if (
                            parent["variable_relatives"][varStr][
                                "complex_positioning"
                            ]
                            == "real"
                        ):
                            realVar = var
                            exhaust1.remove(var)
                            imVar = cousin
                        else:
                            realVar = cousin
                            exhaust1.remove(var)
                            imVar = var
                        holVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][0]
                        antiholVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][1]
                        populate["holVarDict"][holVar] = [realVar, imVar]
                        populate["antiholVarDict"][antiholVar] = [
                            realVar,
                            imVar,
                        ]
                        populate["realVarDict"][realVar] = [holVar, antiholVar]
                        populate["imVarDict"][imVar] = [holVar, antiholVar]
    else:  # _varSpace_type == 'complex'
        for var in varSpace:
            varStr = str(var)
            if var in exhaust1:
                for parent in CVS.values():
                    if varStr in parent["variable_relatives"]:
                        cousin = (
                            set(
                                parent["variable_relatives"][varStr][
                                    "complex_family"
                                ][:2]
                            )
                            - {var}
                        ).pop()
                        if cousin in exhaust1:
                            exhaust1.remove(cousin)
                        if (
                            parent["variable_relatives"][varStr][
                                "complex_positioning"
                            ]
                            == "holomorphic"
                        ):
                            holVar = var
                            exhaust1.remove(var)
                            antiholVar = cousin
                        else:
                            holVar = cousin
                            exhaust1.remove(var)
                            antiholVar = var
                        realVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][2]
                        imVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][3]
                        populate["holVarDict"][holVar] = [realVar, imVar]
                        populate["antiholVarDict"][antiholVar] = [
                            realVar,
                            imVar,
                        ]
                        populate["realVarDict"][realVar] = [holVar, antiholVar]
                        populate["imVarDict"][imVar] = [holVar, antiholVar]
    new_realVarSpace = tuple(populate["realVarDict"].keys())
    new_holVarSpace = tuple(populate["holVarDict"].keys())
    new_antiholVarSpace = tuple(populate["antiholVarDict"].keys())
    new_imVarSpace = tuple(populate["imVarDict"].keys())

    if len(valence) == 0:
        if _varSpace_type == "real":
            populate["realCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["compCoeffDataDict"] = [
                new_holVarSpace + new_antiholVarSpace,
                {(0,) * total_degree: coeff_dict[(0,) * total_degree]},
            ]
        else:
            populate["compCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["realCoeffDataDict"] = [
                new_realVarSpace + new_imVarSpace,
                {(0,) * total_degree: coeff_dict[(0,) * total_degree]},
            ]
    else:

        def _retrieve_indices(term, typeSet=None):
            if typeSet == "symb":
                dictLoc = populate["realVarDict"] | populate["imVarDict"]
                refTuple = new_holVarSpace + new_antiholVarSpace
                termList = dictLoc[term]
            elif typeSet == "real":
                dictLoc = populate["holVarDict"] | populate["antiholVarDict"]
                refTuple = new_realVarSpace + new_imVarSpace
                termList = dictLoc[term]
            index_a = refTuple.index(termList[0])
            index_b = refTuple.index(termList[1], index_a + 1)
            return [index_a, index_b]

        # set up the conversion dicts for index conversion
        if _varSpace_type == "real":
            populate["preProcessMinDataToHol"] = {
                j: _retrieve_indices(varSpace[j], "symb")
                for j in range(len(varSpace))
            }

        else:  # if _varSpace_type == 'complex'
            populate["preProcessMinDataToReal"] = {
                j: _retrieve_indices(varSpace[j], "real")
                for j in range(len(varSpace))
            }

        # coordinate VF and DF conversion
        def decorateWithWeights(index, variance_rule, target="symb"):
            if variance_rule == 0:  # covariant case
                covariance = True
            else:                   # contravariant case
                covariance = False

            if target == "symb":
                if varSpace[index] in variable_registry['conversion_dictionaries']['real_part'].values():
                    holScale = sp.Rational(1, 2) if covariance else 1 # D_z (d_z) coeff of D_x (d_x)
                    antiholScale = sp.Rational(1, 2) if covariance else 1 # D_BARz (d_BARz) coeff of D_x (d_x)
                else:
                    holScale = -I / 2 if covariance else I  # D_z (d_z) coeff of D_y (d_y)
                    antiholScale = I / 2 if covariance else -I  # d_BARz (D_BARz) coeff of d_y (D_y)
                return [
                    [populate["preProcessMinDataToHol"][index][0], holScale],
                    [
                        populate["preProcessMinDataToHol"][index][1],
                        antiholScale,
                    ],
                ]
            else:  # converting from hol to real
                if varSpace[index] in variable_registry['conversion_dictionaries']['holToReal']:
                    realScale = 1 if covariance else sp.Rational(1,2)   # D_x (d_x) coeff in D_z (d_z)
                    imScale = I if covariance else -I*sp.Rational(1,2)  # D_y (d_y) coeff in D_z (d_z)
                else:
                    realScale = 1 if covariance else sp.Rational(1,2)   # D_x (d_x) coeff of D_BARz (d_BARz)
                    imScale = -I if covariance else I*sp.Rational(1,2) # D_y (d_y) coeff of D_BARz (d_BARz)
                return [
                    [populate["preProcessMinDataToReal"][index][0], realScale],
                    [populate["preProcessMinDataToReal"][index][1], imScale],
                ]

        otherDict = dict()
        for term_index, term_coeff in coeff_dict.items():
            if _varSpace_type == "real":
                reformatTarget = "symb"
            else:
                reformatTarget = "real"
            termIndices = [
                decorateWithWeights(k, valence[j], target=reformatTarget) for j,k in enumerate(term_index)
            ]
            prodWithWeights = carProd_with_weights_without_R(*termIndices)
            prodWWRescaled = [[tuple(k[0]), term_coeff * k[1]] for k in prodWithWeights]
            minimal_term_set = _shape_basis(prodWWRescaled,data_shape)
            for term in minimal_term_set:
                if term[0] in otherDict:
                    oldVal = otherDict[term[0]]
                    otherDict[term[0]] = _allToSym(oldVal + term[1])
                else:
                    otherDict[term[0]] = _allToSym(term[1])

        if _varSpace_type == "real":
            populate["realCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["compCoeffDataDict"] = [
                new_holVarSpace + new_antiholVarSpace,
                otherDict,
            ]
        else:
            populate["compCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["realCoeffDataDict"] = [
                new_realVarSpace + new_imVarSpace,
                otherDict,
            ]

    return populate,new_realVarSpace,new_holVarSpace,new_antiholVarSpace,new_imVarSpace

def _shape_basis(basis,shape):
    if shape == 'symmetric':
        old_basis = dict(basis)
        new_basis = dict()
        for index, value in old_basis.items():
            new_index = tuple(sorted(index))
            if new_index in new_basis:
                new_basis[new_index] += value
            else:
                new_basis[new_index] = value
        return list(new_basis.items())
    if shape == 'skew':
        old_basis = dict(basis)
        new_basis = dict()
        for index, value in old_basis.items():
            permS, new_index = permSign(index,returnSorted=True)
            new_index = tuple(new_index)
            if new_index in new_basis:
                new_basis[new_index] += permS*value
            else:
                new_basis[new_index] = permS*value
        return list(new_basis.items())
    return basis


############## clearing and listing
def listVar(
    standard_only=False,
    complex_only=False,
    algebras_only=False,
    zeroForms_only=False,
    coframes_only=False,
    temporary_only=False,
    obscure_only=False,
    protected_only=False,
):
    """
    This function lists all parent labels for objects tracked within the dgcv Variable Management Framework (VMF). In particular strings that are keys in dgcv's internal `standard_variable_systems`, `complex_variable_systems`, 'finite_algebra_systems', 'eps' dictionaries, etc. It also accepts optional keywords to filter the results, showing only temporary, protected, or "obscure" object system labels.

    Parameters
    ----------
    standard_only : bool, optional
        If True, only standard variable system labels will be listed.
    complex_only : bool, optional
        If True, only complex variable system labels will be listed.
    algebras_only : bool, optional
        If True, only finite algebra system labels will be listed.
    zeroForms_only : bool, optional
        If True, only zeroFormAtom system labels will be listed.
    coframes_only : bool, optional
        If True, only coframe system labels will be listed.
    temporary_only : bool, optional
        If True, only variable system labels marked as temporary will be listed.
    protected_only : bool, optional
        If True, only variable system labels marked as protected will be listed.

    Returns
    -------
    list
        A list of object system labels matching the provided filters.

    Notes
    -----
    - If no filters are specified, the function returns all labels tracked in the VMF.
    - If multiple filters are specified, the function combines them, displaying
      labels that meet any of the selected criteria.
    """
    variable_registry = get_variable_registry()

    # Collect all labels
    standard_labels = set(variable_registry["standard_variable_systems"].keys())
    complex_labels = set(variable_registry["complex_variable_systems"].keys())
    algebra_labels = set(variable_registry["finite_algebra_systems"].keys())
    zeroForm_labels = set(variable_registry["eds"]["atoms"].keys())  # New zeroFormAtom labels
    coframe_labels = set(variable_registry["eds"]["coframes"].keys())

    selected_labels = set()
    if standard_only:
        selected_labels |= standard_labels
    if complex_only:
        selected_labels |= complex_labels
    if algebras_only:
        selected_labels |= algebra_labels
    if zeroForms_only:
        selected_labels |= zeroForm_labels
    if coframes_only:
        selected_labels |= coframe_labels

    all_labels = selected_labels if selected_labels else standard_labels | complex_labels | algebra_labels | zeroForm_labels | coframe_labels

    # Apply additional property filters
    if temporary_only:
        all_labels = all_labels & variable_registry["temporary_variables"]
    if obscure_only:
        all_labels = all_labels & variable_registry["obscure_variables"]
    if protected_only:
        all_labels = all_labels & variable_registry["protected_variables"]

    # Return the filtered labels list
    return list(all_labels)

def _clearVar_single(label):
    """
    Helper function that clears a single variable system (standard, complex, or finite algebra)
    from the dgcv variable management framework. Instead of printing a report, it returns
    a tuple (system_type, label) indicating what was cleared.
    """
    registry = get_variable_registry()
    global_vars = _cached_caller_globals
    cleared_info = None

    # If not tracked, nothing to clear
    if label not in registry["_labels"]:
        return None

    path = registry["_labels"][label]["path"]
    branch = path[0]

    # Handle standard variable systems
    if branch == "standard_variable_systems":
        system_dict = registry[branch][label]
        family_names = system_dict["family_names"]
        if isinstance(family_names, str):
            family_names = (family_names,)
        for var in family_names:
            global_vars.pop(var, None)
        global_vars.pop(label, None)
        if system_dict.get("differential_system"):
            for var in family_names:
                global_vars.pop(f"D_{var}", None)
                global_vars.pop(f"d_{var}", None)
            global_vars.pop(f"D_{label}", None)
            global_vars.pop(f"d_{label}", None)
        if system_dict.get("tempVar"):
            registry["temporary_variables"].discard(label)
        if system_dict.get("obsVar"):
            registry["obscure_variables"].discard(label)
        del registry[branch][label]
        cleared_info = ("standard", label)

    # Handle complex variable systems
    elif branch == "complex_variable_systems":
        system_dict = registry[branch][label]
        family_houses = system_dict["family_houses"]
        real_parent, imag_parent = family_houses[-2], family_houses[-1]
        registry["protected_variables"].discard(real_parent)
        registry["protected_variables"].discard(imag_parent)
        if system_dict["family_type"] in ("tuple", "multi_index"):
            for house in family_houses:
                global_vars.pop(house, None)
        variable_relatives = system_dict["variable_relatives"]
        for var_label, var_data in variable_relatives.items():
            global_vars.pop(var_label, None)
            if var_data.get("DFClass"):
                global_vars.pop(f"D_{var_label}", None)
            if var_data.get("VFClass"):
                global_vars.pop(f"d_{var_label}", None)
        conv = registry["conversion_dictionaries"]
        for var_label, var_data in variable_relatives.items():
            pos = var_data.get("complex_positioning")
            val = var_data.get("variable_value")
            if pos == "holomorphic":
                conv["conjugation"].pop(val, None)
                conv["holToReal"].pop(val, None)
                conv["symToReal"].pop(val, None)
            elif pos == "antiholomorphic":
                conv["symToHol"].pop(val, None)
                conv["symToReal"].pop(val, None)
            elif pos in ("real", "imaginary"):
                conv["realToHol"].pop(val, None)
                conv["realToSym"].pop(val, None)
                conv["find_parents"].pop(val, None)
        registry["temporary_variables"].discard(label)
        del registry[branch][label]
        cleared_info = ("complex", label)

    # Handle finite algebra systems
    elif branch == "finite_algebra_systems":
        system_dict = registry[branch][label]
        family_names = system_dict.get("family_names", ())
        for member in family_names:
            global_vars.pop(member, None)
        global_vars.pop(label, None)
        del registry[branch][label]
        cleared_info = ("algebra", label)

    # Handle EDS atoms
    elif branch == "eds" and path[1] == "atoms":
        system_dict = registry["eds"]["atoms"][label]
        family_names = system_dict["family_names"]
        if isinstance(family_names, str):
            family_names = (family_names,)
        for var in family_names:
            global_vars.pop(var, None)
        for var in system_dict.get("family_relatives", {}):
            global_vars.pop(var, None)
        global_vars.pop(label, None)
        del registry["eds"]["atoms"][label]
        cleared_info = ("DFAtom", label)

    # Handle EDS coframes
    elif branch == "eds" and path[1] == "coframes":
        coframe_info = registry["eds"]["coframes"][label]
        cousins_parent = coframe_info.get("cousins_parent")
        global_vars.pop(label, None)
        del registry["eds"]["coframes"][label]
        cleared_info = ("coframe", (label, cousins_parent))

    # Remove from label index
    registry["_labels"].pop(label, None)

    return cleared_info

def clearVar(*labels, report=True):
    """
    Clears variables from the registry and global namespace. Because sometimes, we all need a fresh start.

    This function takes one or more variable system labels (strings) and clears all
    associated variables, vector fields, differential forms, and metadata from the
    dgcv system. Variable system refers to object systems created by the dgcv
    variable creation functions `variableProcedure`, `varWithVF`, and
    `complexVarProc`. Use `listVar()` to retriev a list of existed variable system
    labels. The function handles both standard and complex variable systems,
    ensuring that all related objects are removed from the caller's globals()
    namespace, `variable_registry`, and the conversion dictionaries.

    Parameters
    ----------
    *labels : str
        One or more string labels representing variable systems (either
        standard or complex). These labels will be removed along with all
        associated components.
    report : bool (optional)
        Set True to report about any variable systems cleared from the VMF

    Functionality
    -------------
    - For standard variable systems:
        1. All family members associated with the variable label will be
           removed from the caller's globals() namespace.
        2. If the variable system has associated differential forms (DFClass)
           or vector fields (VFClass), these objects will also be removed.
        3. The label will be removed from `temporary_variables`, if applicable.
        4. Finally, the label will be deleted from `standard_variable_systems`
           in `variable_registry`.

    - For complex variable systems:
        1. For each complex variable system:
            - Labels for the real and imaginary parts will be removed
              from `variable_registry['protected_variables']`.
            - If the system is a tuple, the parent labels for holomorphic,
              antiholomorphic, real, and imaginary variable tuples will be
              removed from the caller's globals() namespace.
            - The `variable_relatives` dictionary will be traversed to remove
              all associated variables, vector fields, and differential forms
              from the caller's globals() namespace.
            - The function will also clean up the corresponding entries in
              `conversion_dictionaries`, depending on the `complex_positioning`
              (holomorphic, antiholomorphic, real, or imaginary).
        2. The complex variable label will be removed from `temporary_variables`,
           if applicable.
        3. Finally, the label will be deleted from `complex_variable_systems`
           in `variable_registry`.

    Notes
    -----
    - Comprehensively clears variables and their associated metadata from the dgcv
      system.
    - Use with `listVar` to expediantly clear everything, e.g., `clearVar(*listVar())`.

    Examples
    --------
    >>> clearVar('x') # removes any dgcv variable system labeled as x, such as
                      # (x, D_x, d_x), (x=(x1, x2), x1, x2, D_x1, d_x1,...), etc.
    >>> clearVar('z', 'y', 'w')

    This will remove all variables, vector fields, and differential forms
    associated with the labels 'z', 'y', and 'w'.

    """
    cleared_standard = []
    cleared_complex = []
    cleared_algebras = []
    cleared_diffFormAtoms = []
    cleared_coframes = []

    for label in labels:
        info = _clearVar_single(label)
        if info:
            system_type, cleared_label = info
            if system_type == "standard":
                cleared_standard.append(cleared_label)
            elif system_type == "complex":
                cleared_complex.append(cleared_label)
            elif system_type == "algebra":
                cleared_algebras.append(cleared_label)
            elif system_type == "DFAtom":
                cleared_diffFormAtoms.append(cleared_label)
            elif system_type == "coframe":
                coframe_label, cousins_parent_label = cleared_label
                cleared_coframes.append((coframe_label, cousins_parent_label))
                clearVar(cousins_parent_label, report=False)

    if report:
        if cleared_standard:
            print(
                f"Cleared standard variable systems from the dgcv variable management framework: {', '.join(cleared_standard)}"
            )
        if cleared_complex:
            print(
                f"Cleared complex variable systems from the dgcv variable management framework: {', '.join(cleared_complex)}"
            )
        if cleared_algebras:
            print(
                f"Cleared finite algebra systems from the dgcv variable management framework: {', '.join(cleared_algebras)}"
            )
        if cleared_diffFormAtoms:
            print(
                f"Cleared differential form systems from the dgcv variable management framework: {', '.join(cleared_diffFormAtoms)}"
            )
        if cleared_coframes:
            for cf_label, cp_label in cleared_coframes:
                print(f"Cleared coframe '{cf_label}' along with associated zero form atom system '{cp_label}'")



############## displaying summaries
def DGCV_snapshot(style=None, use_latex=None, complete_report = None):
    warnings.warn(
        "`DGCV_snapshot` has been deprecated as part of the shift toward standardized naming conventions in the `dgcv` library. "
        "It will be removed in 2026. Please use `vmf_summary` instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return vmf_summary(style=style, use_latex=use_latex, complete_report = complete_report)

def vmf_summary(style=None, use_latex=None, complete_report = None):
    """
    Generate a comprehensive snapshot of dgcv's variable management framework (VMF), such as coordinate systems, algebras, coframes, and more.

    Parameters
    ----------
    style : str, optional
        The style options to apply to the summary table. Default theme is 'chalkboard_green'. Use the dgcv function `get_DGCV_themes()` to display a list of other available themes.
    use_latex : bool, optional
        If True, the table will format text using LaTeX for better
        mathematical display. Default is False.

    Returns
    -------
    DataFrame
        A formatted summary table displaying the initialized objects, such as coordinate systems, algebras, coframes, etc.
    """
    if style is None:
        style = get_dgcv_settings_registry()['theme']
    if use_latex is None:
        use_latex = get_dgcv_settings_registry()['use_latex']

    if complete_report is True:
        force_report = True
    else:
        force_report = False
    if complete_report is None:
        complete_report = True

    vr = get_variable_registry()
    outputs = []

    if vr['standard_variable_systems'] or vr['complex_variable_systems'] or force_report:
        outputs.append(_snapshot_coor_(style=style, use_latex=use_latex))
    if vr['finite_algebra_systems'] or force_report:
        outputs.append(_snapshot_algebras_(style=style, use_latex=use_latex))
    if vr['eds']['atoms'] or force_report:
        outputs.append(_snapshot_eds_atoms_(style=style, use_latex=use_latex))
    if vr['eds']['coframes'] or force_report:
        outputs.append(_snapshot_coframes_(style=style, use_latex=use_latex))

    if not outputs and not force_report:
        print("There are no objects currently registered in the dgcv VMF.")
    else:
        for table in outputs:
            display(table)

def variableSummary(*args, **kwargs):
    warnings.warn(
        "variableSummary() is deprecated and will be removed in a future version. "
        "Please use vmf_summary() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return vmf_summary(*args, **kwargs)

def _snapshot_coor_(style=None, use_latex=None):
    """
    Returns a Pandas DataFrame summarizing the coordinate systems in dgcv.

    This snapshot includes both complex and standard variable systems.
    For each system, the summary includes:
      - The number of variables (tuple length)
      - The initial index
      - For complex systems, the labels used for the real and imaginary parts,
        as well as formatted representations of the vector fields and differential forms.
      - For standard systems, the corresponding formatted vector fields and differential forms.

    Parameters:
        use_latex : bool, optional
            If True, formatting is applied using LaTeX.
        style : str, optional
            A style theme to apply to the summary table.

    Returns:
        A styled Pandas DataFrame summarizing the coordinate systems and finite algebras.
    """
    if style is None:
        style = get_dgcv_settings_registry()['theme']
    if use_latex is None:
        use_latex = get_dgcv_settings_registry()['use_latex']

    def convert_to_greek(var_name):
        for name, greek in greek_letters.items():
            if var_name.lower().startswith(name):
                return var_name.replace(name, greek, 1)
        return var_name

    def format_variable_name(var_name, system_type, use_latex=False):
        if system_type == "standard":
            info = variable_registry["standard_variable_systems"].get(var_name, {})
            family_names = info.get("family_names", var_name)
            initial_index = info.get("initial_index", 1)
        elif system_type == "complex":
            info = variable_registry["complex_variable_systems"].get(var_name, {})
            # For complex systems, assume family_names is a 4-tuple; we use its first element (holomorphic names)
            family_names = info.get("family_names", ())
            if family_names and isinstance(family_names, (list, tuple)) and len(family_names) > 0:
                family_names = family_names[0]
            else:
                family_names = var_name
            initial_index = info.get("initial_index", 1)
        elif system_type == "algebra":
            info = variable_registry["finite_algebra_systems"].get(var_name, {})
            family_names = info.get("family_names", var_name)
            initial_index = None
        else:
            family_names, initial_index = var_name, None

        # If family_names is a sequence with more than one element, format with an ellipsis.
        if isinstance(family_names, (list, tuple)) and len(family_names) > 1:
            if use_latex:
                if initial_index is not None and isinstance(family_names, (list, tuple)) and len(family_names) > 1:
                    content = (f"{format_latex_subscripts(var_name)} = \\left( {format_latex_subscripts(var_name,nest_braces=True)}_{{{initial_index}}}, "
                               f"\\ldots, {format_latex_subscripts(var_name,nest_braces=True)}_{{{initial_index + len(family_names) - 1}}} \\right)")
                else:
                    content = f"{convert_to_greek(var_name)} = {convert_to_greek(var_name)}"
            else:
                content = f"{var_name} = ({family_names[0]}, ..., {family_names[-1]})"
        else:
            content = f"{var_name}"
        if use_latex:
            content = f"${content}$"
        return content

    def format_latex_subscripts(var_name, nest_braces=False):
        """for use_latex branches"""
        if var_name[-1]=='_':
            var_name=var_name[:-1]
        parts = var_name.split("_")
        if len(parts) == 1:
            return convert_to_greek(var_name)
        base = convert_to_greek(parts[0])
        subscript = ", ".join(parts[1:])
        if nest_braces is True:
            return f"{{{base}_{{{subscript}}}}}"
        else:
            return f"{base}_{{{subscript}}}"

    def build_object_string(obj_type, var_name, start_index, tuple_len, system_type, use_latex=False):
        # For standard systems, if there's only one variable, return a simple label;
        # otherwise, show the range of indices.
        if tuple_len == 1:
            if use_latex:
                if use_latex:
                    var_name=format_latex_subscripts(var_name)
                if obj_type=='D':
                    content = f"$\\frac{{\\partial}}{{\\partial {var_name}}}$"
                else:
                    content = f"$\\operatorname{{d}} {var_name}$"
            else:
                content = f"{obj_type}_{var_name}"
        else:
            if use_latex:
                if use_latex:
                    var_name=format_latex_subscripts(var_name,nest_braces=True)
                if obj_type=='D':
                    content = f"$\\frac{{\\partial}}{{\\partial {var_name}_{{{start_index}}}}}$, $\\ldots$, $\\frac{{\\partial}}{{\\partial {var_name}_{{{start_index + tuple_len - 1}}}}}$"
                else:
                    content = f"$\\operatorname{{d}} {var_name}_{{{start_index}}}$, $\\ldots$, $\\operatorname{{d}} {var_name}_{{{start_index + tuple_len - 1}}}$"
            else:
                content = f"{obj_type}_{var_name}{start_index},...,{obj_type}_{var_name}{start_index + tuple_len - 1}"
        return content

    def build_object_string_for_complex(obj_type, family_houses, family_names, start_index, use_latex=False):
        """
        For a complex variable system, family_houses is expected to be a 4-tuple of labels
        (e.g. (hol, anti, real, imag)) and family_names is a 4-tuple of sequences, one for each part.
        This function formats each part (only if the corresponding list has more than one element)
        using an ellipsis; otherwise, it shows the single variable.
        """
        parts = []
        # Expect family_names to be a 4-tuple: (hol_names, anti_names, real_names, imag_names)
        if isinstance(family_names, (list, tuple)) and len(family_names) == 4:
            for i, part in enumerate(family_houses):
                part_names = family_names[i]  # list of names for this part
                if isinstance(part_names, (list, tuple)) and len(part_names) > 1:
                    if use_latex:
                        part_str = f"\\frac{{\\partial}}{{\\partial {format_latex_subscripts(part,nest_braces=True)}_{{{start_index}}}}}, \\ldots, \\frac{{\\partial}}{{\\partial {format_latex_subscripts(part,nest_braces=True)}_{{{start_index + len(part_names) - 1}}}}}" if obj_type=="D" else f"d {format_latex_subscripts(part,nest_braces=True)}_{{{start_index}}}, \\ldots, d {format_latex_subscripts(part,nest_braces=True)}_{{{start_index + len(part_names) - 1}}}"
                        part_str = f"${part_str}$"
                    else:
                        part_str = f"{obj_type}_{part}{start_index},...,{obj_type}_{part}{start_index + len(part_names) - 1}"
                else:
                    if use_latex:
                        part_str = f"${obj_type}_{format_latex_subscripts(part)}$"
                    else:
                        part_str = f"{obj_type}_{part}"
                parts.append(part_str)
        else:
            # Fallback: if family_names is not the expected 4-tuple, treat as a single list.
            if isinstance(family_names, (list, tuple)) and len(family_names) > 1:
                if use_latex:
                    part_str = f"${obj_type}_{format_latex_subscripts(family_houses[0])}{start_index}$, $\\ldots$, ${obj_type}_{format_latex_subscripts(family_houses[0])}{start_index + len(family_names) - 1}$"
                else:
                    part_str = f"{obj_type}_{family_houses[0]}{start_index},...,{obj_type}_{family_houses[0]}{start_index + len(family_names) - 1}"
                parts.append(part_str)
            else:
                if use_latex:
                    parts.append(f"${obj_type}_{format_latex_subscripts(family_houses[0])}$")
                else:
                    parts.append(f"{obj_type}_{family_houses[0]}")
        return ", ".join(parts)

    variable_registry = get_variable_registry()
    data = []          # Each row is: [# Variables, Initial Index, Real Part, Imaginary Part, Vector Fields, Differential Forms]
    var_system_labels = []  # Regular column for variable system labels

    # Process complex variable systems
    for var_name in sorted(variable_registry.get("complex_variable_systems", {}).keys()):
        system = variable_registry["complex_variable_systems"][var_name]
        # For complex systems, family_names should be a 4-tuple.
        fn = system.get("family_names", ())
        if fn and isinstance(fn, (list, tuple)) and len(fn) == 4:
            hol_names = fn[0]
        else:
            hol_names = []
        tuple_len = len(hol_names) if isinstance(hol_names, (list, tuple)) else 1
        start_index = system.get("initial_index", 1)
        formatted_label = format_variable_name(var_name, "complex", use_latex=use_latex)
        var_system_labels.append(formatted_label)
        family_houses = system.get("family_houses", ("N/A", "N/A", "N/A", "N/A"))
        # For complex systems, display real and imaginary parts from the 3rd and 4th entries.
        if isinstance(fn, (list, tuple)) and len(fn)==4:
            real_names = fn[2]
            imag_names = fn[3]
        else:
            real_names, imag_names = "N/A", "N/A"
        if use_latex:
            real_part = f"${format_latex_subscripts(family_houses[2])} = \\left( {format_latex_subscripts(family_houses[2],nest_braces=True)}_{{{start_index}}}, \\ldots, {format_latex_subscripts(family_houses[2],nest_braces=True)}_{{{start_index + len(real_names) - 1}}} \\right)$" if isinstance(real_names, (list, tuple)) and len(real_names)>1 else f"${format_latex_subscripts(family_houses[2])}$"
            imag_part = f"${format_latex_subscripts(family_houses[3])} = \\left( {format_latex_subscripts(family_houses[3],nest_braces=True)}_{{{start_index}}}, \\ldots, {format_latex_subscripts(family_houses[3],nest_braces=True)}_{{{start_index + len(imag_names) - 1}}} \\right)$" if isinstance(imag_names, (list, tuple)) and len(imag_names)>1 else f"${format_latex_subscripts(family_houses[3])}$"
        else:
            real_part = f"{family_houses[2]} = ({real_names[0]}, ..., {real_names[-1]})" if isinstance(real_names, (list, tuple)) and len(real_names)>1 else f"{family_houses[2]}"
            imag_part = f"{family_houses[3]} = ({imag_names[0]}, ..., {imag_names[-1]})" if isinstance(imag_names, (list, tuple)) and len(imag_names)>1 else f"{family_houses[3]}"
        vf_str = build_object_string_for_complex("D", family_houses, fn, start_index, use_latex)
        df_str = build_object_string_for_complex("d", family_houses, fn, start_index, use_latex)
        data.append([tuple_len, real_part, imag_part, vf_str, df_str])

    # Process standard variable systems
    for var_name in sorted(variable_registry.get("standard_variable_systems", {}).keys()):
        system = variable_registry["standard_variable_systems"][var_name]
        family_names = system.get("family_names", ())
        tuple_len = len(family_names) if isinstance(family_names, (list, tuple)) else 1
        start_index = system.get("initial_index", 1)
        formatted_label = format_variable_name(var_name, "standard", use_latex=use_latex)
        var_system_labels.append(formatted_label)
        vf_str = build_object_string("D", var_name, start_index, tuple_len, "standard", use_latex)
        df_str = build_object_string("d", var_name, start_index, tuple_len, "standard", use_latex)
        data.append([tuple_len, "----", "----", vf_str, df_str])


    # Combine labels from coordinate systems and algebras.
    all_labels = var_system_labels
    # Prepend the label to each data row.
    combined_data = []
    for label, row in zip(all_labels, data):
        combined_data.append([label] + row)

    columns = ["Coordinate System", "# of Variables", "Real Part", "Imaginary Part", "Vector Fields", "Differential Forms"]
    table = pd.DataFrame(combined_data, columns=columns)
    table_header = "Initialized Coordinate Systems"
    table_styles = get_style(style)+[{'selector': 'td', 'props': [('text-align', 'left')]}]+[{'selector': 'th', 'props': [('text-align', 'left')]}]
    styled_table = (
        table.style
        .set_table_styles(table_styles)
        .set_caption(f"{table_header}")
        .hide(axis="index")
    )
    return styled_table

def _snapshot_algebras_(style=None, use_latex=None):
    """
    Returns a Pandas DataFrame summarizing the finite algebra systems in dgcv's VMF.

    For each finite algebra system, the snapshot includes:
      - The formatted algebra label.
      - The family type.
      - The basis (family_names), formatted as a comma‑separated string (with ellipsis if there are too many elements).
      - The number of basis elements.

    Args:
        use_latex (bool, optional): If True, formats labels using LaTeX.
        style (str, optional): A style theme name to retrieve table styles.

    Returns:
        A styled Pandas DataFrame.
    """
    if style is None:
        style = get_dgcv_settings_registry()['theme']
    if use_latex is None:
        use_latex = get_dgcv_settings_registry()['use_latex']
    def convert_to_greek(var_name):
        # Assume greek_letters is defined module-wide as a dict mapping strings to LaTeX-valid Greek strings.
        # Example: {'alpha': '\\alpha', 'beta': '\\beta', ...}
        for name, greek in greek_letters.items():
            if var_name.lower().startswith(name):
                return var_name.replace(name, greek, 1)
        return var_name

    def process_basis_label(label, ul):
        if ul is True:
            # Use regex to separate alphabetic and numeric parts.
            match = re.match(r"(.*?)(\d+)?$", label)
            if match:
                basis_elem_name = match.group(1).replace("_", "")
                basis_elem_number = match.group(2)
                basis_elem_name = convert_to_greek(basis_elem_name)
                if basis_elem_number:
                    return f"${basis_elem_name}_{{{basis_elem_number}}}$"
                else:
                    return f"${basis_elem_name}$"
            else:
                return f'${label}$ '
        else:
            return label

    registry = get_variable_registry()
    data = []

    finite_algebras = registry.get("finite_algebra_systems", {})
    for label in sorted(finite_algebras.keys()):
        system = finite_algebras[label]
        family_names = system.get("family_names", ())
        if isinstance(family_names, (list, tuple)):
            num_basis = len(family_names)
            if num_basis > 5:
                basis_str = f"{process_basis_label(family_names[0],ul=use_latex)}, ..., {process_basis_label(family_names[-1],ul=use_latex)}"
            else:
                basis_str = ", ".join(process_basis_label(x,ul=use_latex) for x in family_names)
        else:
            num_basis = 1
            basis_str = process_basis_label(family_names,ul=use_latex)
        # For the algebra label, use the module-wide format_variable_name with system_type "algebra".
        if use_latex and label in _cached_caller_globals:
            formatted_label = _cached_caller_globals[label]._repr_latex_(abbrev=True)
        else:
            formatted_label = label

        # Add grading column
        if label in _cached_caller_globals:
            alg = _cached_caller_globals[label]
            grading = alg.grading
            if isinstance(grading, (list, tuple)) and len(grading) > 0 and all(isinstance(g, (list, tuple)) for g in grading) and any(g for g in grading):
                grading_str = ", ".join(f"({', '.join(map(str, g))})" for g in grading)
            else:
                grading_str = "None"
        else:
            grading_str = "None"

        data.append([formatted_label, basis_str, num_basis, grading_str])

    columns = ["Algebra Label", "Basis", "Dimension", "Grading"]
    df = pd.DataFrame(data, columns=columns)

    table_styles = get_style(style)+[{'selector': 'td', 'props': [('text-align', 'left')]}]+[{'selector': 'th', 'props': [('text-align', 'left')]}]
    styled_df = df.style.set_table_styles(table_styles)
    styled_df = (
        df.style
        .set_table_styles(table_styles)
        .set_caption("Initialized Finite-dimensional Algebras")
        .hide(axis="index")
    )
    return styled_df

def _snapshot_eds_atoms_(style=None, use_latex=None):
    """
    Returns a summary table listing abstract differential form atoms in the VMF scope

    Parameters:
    ----------
    style : str, optional
        A style theme to apply to the summary table.
    use_latex : bool, optional
        If True, formats text using LaTeX

    Returns:
    -------
    pandas.DataFrame
        A summary table listing abstract differential form atoms in the VMF scope
    """
    if style is None:
        style = get_dgcv_settings_registry()['theme']
    if use_latex is None:
        use_latex = get_dgcv_settings_registry()['use_latex']
    variable_registry = get_variable_registry()
    eds_atoms_registry = variable_registry["eds"]["atoms"]

    if not eds_atoms_registry:
        return pd.DataFrame(columns=["DF System", "Degree", "# Elements", "Differential Forms", "Conjugate Forms", "Primary Coframe"])

    data = []
    for label, system in sorted(eds_atoms_registry.items()):
        df_system = label
        degree = system.get("degree", "----")
        family_values = system.get("family_values", ())
        num_elements = len(family_values) if isinstance(family_values, tuple) else 1

        # Format Differential Forms
        if isinstance(family_values, tuple):
            if len(family_values) > 3:
                diff_forms = f"{family_values[0]}, ..., {family_values[-1]}"
            else:
                diff_forms = ", ".join(str(x) for x in family_values)
        else:
            diff_forms = str(family_values)

        # Apply LaTeX formatting for Differential Forms if required
        if use_latex and family_values:
            if isinstance(family_values, tuple) and len(family_values) > 3:
                diff_forms = f"$ {family_values[0]._latex()}, ..., {family_values[-1]._latex()} $"
            elif isinstance(family_values, tuple):
                diff_forms = ", ".join(f"{x._latex()}" if hasattr(x, "_latex") else str(x) for x in family_values)
                diff_forms = f"$ {diff_forms} $"

        # Format Conjugate Forms
        real_status = system.get("real", False)
        if real_status:
            conjugate_forms = "----"
        else:
            conjugates = system.get("conjugates", {})
            if conjugates:
                conjugate_list = list(conjugates.values())
                if len(conjugate_list) > 3:
                    conjugate_forms = f"{conjugate_list[0]}, ..., {conjugate_list[-1]}"
                else:
                    conjugate_forms = ", ".join(str(x) for x in conjugate_list)
                if use_latex and conjugate_list and hasattr(conjugate_list[0], "_latex"):
                    if len(conjugate_list) > 3:
                        conjugate_forms = f"$ {conjugate_list[0]._latex()}, ..., {conjugate_list[-1]._latex()} $"
                    else:
                        conjugate_forms = ", ".join(x._latex() for x in conjugate_list)
                        conjugate_forms = f"$ {conjugate_forms} $"
            else:
                conjugate_forms = "----"

        # Format Primary Coframe
        primary_coframe = system.get("primary_coframe", None)
        if primary_coframe is None:
            primary_coframe_str = "----"
        else:
            primary_coframe_str = primary_coframe._latex() if use_latex and hasattr(primary_coframe, "_latex") else repr(primary_coframe)

        data.append([df_system, degree, num_elements, diff_forms, conjugate_forms, primary_coframe_str])

    columns = ["DF System", "Degree", "# Elements", "Differential Forms", "Conjugate Forms", "Primary Coframe"]
    df = pd.DataFrame(data, columns=columns)

    table_styles = get_style(style) + [
        {'selector': 'th', 'props': [('text-align', 'left')]},
        {'selector': 'td', 'props': [('text-align', 'left')]}
    ]
    styled_df = df.style.set_table_styles(table_styles).set_caption("Initialized abstract differential forms in the VMF scope").hide(axis="index")

    return styled_df

def _snapshot_coframes_(style=None, use_latex=None):
    """
    Returns a summary table listing coframes in the VMF scope

    Parameters:
    ----------
    style : str, optional
        A style theme to apply to the summary table.
    use_latex : bool, optional
        If True, formats text using LaTeX

    Returns:
    -------
    pandas.DataFrame
    Returns a summary table listing coframes in the VMF scope
    """

    if style is None:
        style = get_dgcv_settings_registry()['theme']
    if use_latex is None:
        use_latex = get_dgcv_settings_registry()['use_latex']
    variable_registry = get_variable_registry()
    coframes_registry = variable_registry["eds"].get("coframes", {})

    if not coframes_registry:
        return pd.DataFrame(columns=["Coframe Label", "Coframe 1-Forms", "Structure Coefficients"])

    data = []
    for label, system in sorted(coframes_registry.items()):
        # Retrieve coframe object
        coframe_obj = _cached_caller_globals.get(label, label)
        coframe_label = label

        # Process Coframe 1-Forms
        if isinstance(coframe_obj,str):
            children = system.get("children", [])
            if len(children)>0:
                if any(child not in _cached_caller_globals for child in children):
                    children = []
                else:
                    children = [_cached_caller_globals[child] for child in children]
        else:
            children = coframe_obj.forms
        if len(children) > 3:
            coframe_1_forms = f"{children[0]._latex()}, ..., {children[-1]._latex()}"
        else:
            coframe_1_forms = ", ".join(
                child._latex() if use_latex and hasattr(child, "_latex") else repr(child)
                for child in children
            )
        coframe_1_forms = f"$ {coframe_1_forms} $" if use_latex else coframe_1_forms

        # Process Structure Coefficients
        cousins = system.get("cousins_vals", [])
        if len(cousins) > 3:
            structure_coeffs = f"{cousins[0]._latex()}, ..., {cousins[-1]._latex()}"
        else:
            structure_coeffs = ", ".join(
                cousin._latex() if use_latex and hasattr(cousin, "_latex") else repr(cousin)
                for cousin in cousins
            )
        structure_coeffs = f"$ {structure_coeffs} $" if use_latex else structure_coeffs

        data.append([coframe_label, coframe_1_forms, structure_coeffs])

    columns = ["Coframe Label", "Coframe 1-Forms", "Structure Coefficients"]
    df = pd.DataFrame(data, columns=columns)

    # Apply styling
    table_styles = get_style(style) + [
        {'selector': 'th', 'props': [('text-align', 'left')]},
        {'selector': 'td', 'props': [('text-align', 'left')]}
    ]
    styled_df = df.style.set_table_styles(table_styles).set_caption("Initialized Abstract Coframes").hide(axis="index")

    return styled_df




import warnings

import sympy as sp

from .._config import get_dgcv_settings_registry, get_variable_registry
from .._safeguards import (
    _cached_caller_globals,
    get_dgcv_category,
    retrieve_passkey,
    retrieve_public_key,
    validate_label,
    validate_label_list,
)
from ..combinatorics import carProd
from ..dgcv_core import variableProcedure
from ..solvers import solve_dgcv
from ..tensors import tensorProduct
from ..vmf import clearVar, listVar
from .algebras_aux import _validate_structure_data
from .algebras_core import (
    algebra_class,
    algebra_element_class,
    algebra_subspace_class,
)


class subalgebra_class(algebra_subspace_class):
    def __new__(cls, basis, alg, grading=None, _compressed_structure_data=None, _internal_lock=None):
        return super().__new__(cls, basis, alg, test_weights=None, _grading=grading, _internal_lock=_internal_lock)
    def __init__(self, basis, alg, grading=None, _compressed_structure_data=None, _internal_lock=None):
        super().__init__(basis, alg, test_weights=None, _grading=grading, _internal_lock=_internal_lock)
        basis = self.filtered_basis
        self.structureData = None
        if _internal_lock==retrieve_passkey():
            if _compressed_structure_data is not None:
                self.structureData=_compressed_structure_data
        if self.structureData is None:
            self.structureData = self.is_subalgebra(return_structure_data=True)['structure_data']
        self._structureData=tuple(map(tuple, self.structureData))
        self.subindices_to_ambiant_dict = {count:elem for count,elem in enumerate(basis)}
        self.basis_in_ambiant_alg = tuple(basis)
        self.basis = [subalgebra_element(self,[1 if j==count else 0 for j in range(self.dimension)],elem.valence) for count,elem in enumerate(basis)]
        if all(elem in self.ambiant for elem in basis):
            self.basis_labels = [elem.__str__() for elem in self.basis]
        else:
            self.basis_labels = [f'_e_{j+1}' for j in range(self.dimension)]
        self._dgcv_class_check=retrieve_passkey()
        self._dgcv_category='subalgebra'

# (self,alg,coeffs,valence,ambiant_rep=None,_internalLock=None)
    def _class_builder(self,coeffs,valence):
        return subalgebra_element(self,coeffs,valence)

    def __eq__(self, other):
        if not isinstance(other, subalgebra_class):
            return NotImplemented
        return (self.basis_in_ambiant_alg == other.basis_in_ambiant_alg)

    def __hash__(self):
        return hash(self.basis_in_ambiant_alg)


    def multiplication_table(self, elements=None, restrict_to_subspace=False, style=None, use_latex=None):
        if elements is None:
            newElements = [elem.ambiant_rep for elem in self.basis]
        elif isinstance(elements,(list,tuple)):
            warningMessage = ''
            newElements=[]
            for elem in elements:
                elemTest = elem.ambiant_rep if isinstance(elem,subalgebra_element) else elem
                if self.contains(elemTest) is False:
                    if warningMessage=='':
                        warningMessage+='Some elements in the `elements` list were not in the span of the subalgebra\'s basis, so they were omitted from the multiplication table.'
                else:
                    newElements.append(elem)
            if warningMessage!=0 and len(newElements)>0:
                warnings.warn(warningMessage)
            else:
                raise TypeError('No elements from the provided `elements` list belong to the subalgebra, so a multiplication table will not be produced.') from None
        else:
            raise TypeError('If provided, the `elements` parameter in `subalgebra_class.multiplication_table` must be a list.') from None

        return self.ambiant.multiplication_table(elements=newElements, restrict_to_subspace=restrict_to_subspace, style=style, use_latex=use_latex,_called_from_subalgebra={'internalLock':retrieve_passkey(),'basis':self.basis})

    def subalgebra(self,basis,grading=None):
        return self.ambiant.subalegra(basis,grading=grading)
    def subspace(self,basis,grading=None):
        return self.ambiant.subspace(basis,grading=grading)
    def contains(self, items, return_basis_coeffs = False):
        if not isinstance(items,(list,tuple)):
            items = [items]
        for item in items:
            if get_dgcv_category(item)=='subalgebra_element':
                if item.algebra==self:
                    bas=self.basis
                elif item.algebra.ambiant==self.ambiant:
                    item=item.ambiant_rep
                    bas=self.basis_in_ambiant_alg
                else:
                    return False
            elif get_dgcv_category(item)=='algebra_element' and item.algebra.ambiant==self.ambiant:
                item=item.ambiant_rep
                bas=self.basis_in_ambiant_alg
            else:
                return False
            if item not in bas:
                tempVarLabel = "T" + retrieve_public_key()
                variableProcedure(tempVarLabel, len(bas), _tempVar=retrieve_passkey())
                genElement = sum([_cached_caller_globals[tempVarLabel][j+1] * elem for j,elem in enumerate(bas[1:])],_cached_caller_globals[tempVarLabel][0]*(bas[0]))
                sol = solve_dgcv(item-genElement,_cached_caller_globals[tempVarLabel])
                if len(sol)==0:
                    clearVar(*listVar(temporary_only=True))
                    return False
            else:
                if return_basis_coeffs is True:
                    idx = bas.index(item)
                    return [1 if _==idx else 0 for _ in range(len(bas))]
        if return_basis_coeffs is True:
            vec=[var.subs(sol[0]) for var in _cached_caller_globals[tempVarLabel]]
            clearVar(*listVar(temporary_only=True))
            return vec
        clearVar(*listVar(temporary_only=True))
        return True

class subalgebra_element():
    def __init__(self,alg,coeffs,valence,ambiant_rep=None,_internalLock=None):
        self.algebra = alg
        self.vectorSpace = alg
        self.valence = valence
        if isinstance(coeffs, (list, tuple)):  
            self.coeffs = tuple(coeffs)
        else:
            raise TypeError("subalgebra_element expects coeffs to be a list or tuple.") from None
        if _internalLock==retrieve_passkey():
            self._ambiant_rep = ambiant_rep
        else:
            self._ambiant_rep = None
        self._dgcv_class_check=retrieve_passkey()
        self._dgcv_category='subalgebra_element'

    @property
    def ambiant_rep(self):
        if self._ambiant_rep is None:
            self._ambiant_rep = sum([coeff*self.algebra.subindices_to_ambiant_dict[j+1] for j,coeff in enumerate(self.coeffs[1:])],self.coeffs[0]*self.algebra.subindices_to_ambiant_dict[0])
        return self._ambiant_rep

    def __eq__(self, other):
        if not isinstance(other, subalgebra_element):
            return NotImplemented
        return (self.algebra == other.algebra and self.coeffs == other.coeffs and self.valence == other.valence)
    def __hash__(self):
        return hash((self.algebra, self.coeffs, self.valence))

    def __contains__(self, item):
        return item in self.basis

    def __iter__(self):
        return iter(self.basis) 

    def __getitem__(self, indices):
        if isinstance(indices,int):
            return self.basis[indices]
        elif isinstance(indices,list): 
            if len(indices)==1:
                return self.basis[indices[0]]
            elif isinstance(indices,list) and len(indices)==2:
                return self.structureData[indices[0]][indices[1]]
            elif isinstance(indices,list) and len(indices)==3:
                return self.structureData[indices[0]][indices[1]][indices[2]]
        else:
            raise TypeError(f'To access a subalgebra element or structure data component, provide one index for an element from the basis, two indices for a list of coefficients from the product  of two basis elements, or 3 indices for the corresponding entry in the structure array. Instead of an integer of list of integers, the following was given: {indices}') from None



    def __str__(self):
        return self.ambiant_rep.__str__()
    def _repr_latex_(self,verbose=False):
        return self.ambiant_rep._repr_latex_(verbose=verbose)
    def _latex(self,printer=None):
        return self._repr_latex_()
    def _sympystr(self):
        return self.ambiant_rep._sympystr()
    def _latex_verbose(self, printer=None):
        return self.ambiant_rep._latex_verbose(printer=printer)
    def __repr__(self):
        return self.ambiant_rep.__repr__()

    def is_zero(self):
        for j in self.coeffs:
            if sp.simplify(j) != 0:
                return False
        else:
            return True

    def subs(self, subsData):
        newCoeffs = [sp.sympify(j).subs(subsData) for j in self.coeffs]
        return subalgebra_element(self.algebra, newCoeffs, self.valence)

    def dual(self):
        return subalgebra_element(self.algebra, self.coeffs, (self.valence+1)%2)

    def _convert_to_tp(self):
        return tensorProduct(self.algebra,{(j,self.valence):self.coeffs[j] for j in range(self.algebra.dimension)})

    def _recursion_contract_hom(self, other):
        return self._convert_to_tp()._recursion_contract_hom(other)

    def __add__(self, other):
        if hasattr(other,"is_zero") and other.is_zero():
            return self
        if isinstance(other, subalgebra_element):
            if self.algebra == other.algebra and self.valence==other.valence:
                return subalgebra_element(
                    self.algebra,
                    [self.coeffs[j] + other.coeffs[j] for j in range(len(self.coeffs))],
                    self.valence,
                )
            else:
                raise TypeError(
                    "subalgebra_element operands for + must belong to the same subalgebra."
                ) from None
        return self.ambiant_rep.__add__(other)

    def __sub__(self, other):
        if hasattr(other,"is_zero") and other.is_zero():
            return self
        if isinstance(other, subalgebra_element):
            if self.algebra == other.algebra and self.valence==other.valence:
                return subalgebra_element(
                    self.algebra,
                    [self.coeffs[j] - other.coeffs[j] for j in range(len(self.coeffs))],
                    self.valence,
                )
            else:
                raise TypeError(
                    "subalgebra_element operands for - must belong to the same algebra."
                ) from None
        return self.ambiant_rep.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, subalgebra_element):
            if self.algebra == other.algebra and self.valence==other.valence:
                result_coeffs = [0] * self.algebra.dimension
                for i in range(self.algebra.dimension):
                    for j in range(self.algebra.dimension):
                        scalar_product = self.coeffs[i] * other.coeffs[j]
                        structure_vector_product = [
                            scalar_product * element
                            for element in self.algebra.structureData[i][j]
                        ]
                        result_coeffs = [
                            sp.sympify(result_coeffs[k] + structure_vector_product[k])
                            for k in range(len(result_coeffs))
                        ]
                return subalgebra_element(self.algebra, result_coeffs, self.valence)
            else:
                raise TypeError(
                    "Both operands for * must be subalgebra_element instances from the same algebra."
                ) from None
        elif isinstance(other, (int, float, sp.Expr)):
            new_coeffs = [coeff * other for coeff in self.coeffs]
            return subalgebra_element(self.algebra, new_coeffs, self.valence)
        elif isinstance(other,tensorProduct):
            return other.__rmul__(self)
        return self.ambiant_rep.__mul__(other)

    def __rmul__(self, other):
        if isinstance(
            other, (int, float, sp.Expr)
        ):
            return self * other
        elif isinstance(other, subalgebra_element):
            other * self
        elif isinstance(other,tensorProduct):
            return other.__mul__(self)
        return self.ambiant_rep.__rmul__(other)

    def __matmul__(self, other):
        """Overload @ operator for tensor product."""
        if isinstance(other, subalgebra_element) and other.algebra==self.algebra:
            new_dict = {(j,k,self.valence,other.valence):self.coeffs[j]*other.coeffs[k] for j in range(self.algebra.dimension) for k in range(self.algebra.dimension)}
            return tensorProduct(self.algebra, new_dict)
        return self.ambiant_rep.__matmul__(other)

    def __xor__(self, other):
        if other == '':
            return self.dual()
        raise ValueError("Invalid operation. Use `^''` to denote the dual.") from None

    def __neg__(self):
        return -1*self

    def check_element_weight(self):
        """
        Determines the weight vector of this subalgebra_element with respect to its ambiant algebra's grading vectors.

        Returns
        -------
        list
            A list of weights corresponding to the grading vectors of the parent algebra.
            Each entry is either an integer, sympy.Expr (weight), the string 'AllW' if the element is the zero element,
            or 'NoW' if the element is not homogeneous.

        Notes
        -----
        - This method calls the parent algebra' check_element_weight method.
        - 'AllW' is returned for zero elements, which are compaible with all weights.
        - 'NoW' is returned for non-homogeneous elements that do not satisfy the grading constraints.
        """
        return self.algebra.check_element_weight(self)

class simpleLieAlgebra(algebra_class):
    def __init__(self, structure_data, grading=None, format_sparse=False, process_matrix_rep=False, preferred_representation=None, _label=None, _basis_labels=None, _calledFromCreator=None, _callLock=None, _print_warning=None, _child_print_warning=None, _exclude_from_VMF=None,_simple_data=None):
        if _calledFromCreator!=retrieve_passkey():
            raise RuntimeError('`simpleLieAlgebra` class instances can only be initialized by internal `dgcv` functions indirectly.' ) from None
        super().__init__(structure_data, grading, format_sparse, process_matrix_rep, preferred_representation, _label, _basis_labels, _calledFromCreator, _callLock, _print_warning, _child_print_warning, _exclude_from_VMF)

        # assuming grading vectors are complete and given relative to simple roots
        self.roots=[]
        self.simpleRoots=[]
        self.rootSpaces={(0,)*len(self.grading):[]}
        def isSimpleRoot(vec):
            if vec.count(0)==len(vec)-1 and vec.count(1)==1:
                return True
            else:
                return False
        for elem in self.basis:
            root=tuple(elem.check_element_weight())
            if root in self.rootSpaces:
                self.rootSpaces[root].append(elem)
            else:
                self.rootSpaces[root]=[elem]
                self.roots.append(root)
                if isSimpleRoot(root):
                    self.simpleRoots.append(root)
        self.simpleRootSpaces = {root:self.rootSpaces[root] for root in self.simpleRoots}
        seriesLabel,rank = _simple_data['type']
        self.rank = rank
        self.Cartan_subalgebra = self.basis[0:rank]
        self.simpleLieType = f'{seriesLabel}{rank}' # example: "A3", "D4", ... etch

    def root_space_summary(self):
        def pluralize(idx):
            if idx!=1:
                return 's'
            else:
                return ''
        def rootString(idx):
            if idx==1: 
                return '(r_1)'
            if idx==2:
                return '(r_1, r_2)'
            if idx==3:
                return '(r_1, r_2, r_3)'
            else:
                return f'(r_1, ..., r_{idx})'
        print(f'This simple algebra {self.simpleLieType} has {self.rank} root{pluralize(self.rank)} {rootString(self.rank)}, which are dual to the Cartan subalgebra basis {self.Cartan_subalgebra}. These roots correspond to vertices in the Dynkin diagram as follows:\n')

        if self.simpleLieType[0] == 'D':
            n = self.rank
            if n == 2:
                print("Dynkin diagram for D2 is just two disconnected vertices corresponding to a direct sum of two u(2) copies.")
            else:
                lines = []
                horiz = "   "
                if n > 7:
                    mid_nodes = ["r_1 r_2", f"r_{n-4}", f"r_{n-3}", f"r_{n-2}"]
                    latter_rules = [" ◯───◯─"+" ┅ ─","─"*(len(mid_nodes[1])),"─"*(len(mid_nodes[2])),"─"*(len(mid_nodes[3])),""]
                    horiz += "◯".join(latter_rules)
                    top_labels = f"{' '*4}{mid_nodes[0]}{' '*3}{mid_nodes[1]} {mid_nodes[2]} {mid_nodes[3]}"
                    fork_pos = 16+len(latter_rules[1])+len(latter_rules[2])
                elif n>1:
                    horiz += "───".join("◯" for _ in range(n - 2))
                    top_labels = "   " + " ".join([f"r_{i+1}" for i in range(n-2)])
                    horiz += "───◯"
                    fork_pos = 4*(n-2)-1
                else:
                    horiz += "───".join("◯" for _ in range(n - 2))
                    top_labels = "   " + " ".join([f"r_{i+1}" for i in range(n-2)])
                    horiz += "───◯"
                    fork_pos = 4*(n-2)-1

                top_labels += " " + f"r_{n-1}"

                # Final node
                final_line = " " * fork_pos + "│"
                final_node = " " * fork_pos + f"◯ r_{n}"

                # bounding box
                width_bound= len(top_labels)
                title = "│" + self.simpleLieType.center(width_bound) + " │"
                border_top = "┌"+"─"*width_bound+"─┐"
                head_sep = "╞"+"═"*width_bound+"═╡"
                top_labels = "│" + top_labels + " │"
                horiz = "│" + horiz.ljust(width_bound) + " │"
                final_line = "│" + final_line.ljust(width_bound) + " │"
                final_node = "│" + final_node.ljust(width_bound) + " │"
                border_bottom= "└"+"─"*width_bound+"─┘"


                lines.append(border_top)
                lines.append(title)
                lines.append(head_sep)
                lines.append(top_labels)
                lines.append(horiz)
                lines.append(final_line)
                lines.append(final_node)
                lines.append(border_bottom)

                print("\n".join(lines))
        elif self.simpleLieType[0]=='B':
            n = self.rank
            lines = []
            horiz = "   "
            if n > 7:
                mid_nodes = ["r_1 r_2", f"r_{n-3}", f"r_{n-2}", f"r_{n-1}"]
                latter_rules = [" ◯───◯─"+" ⋯ ─","─"*(len(mid_nodes[1])),"─"*(len(mid_nodes[2])),"═"*(len(mid_nodes[3])-2)+"═>",""]
                horiz += "◯".join(latter_rules)
                top_labels = f"{' '*4}{mid_nodes[0]}{' '*3}{mid_nodes[1]} {mid_nodes[2]} {mid_nodes[3]}"
            else:
                horiz += "───".join("◯" for _ in range(n-1))
                top_labels = "   " + " ".join([f"r_{i+1}" for i in range(n-1)])
                horiz += "══>◯"
            top_labels += " " + f"r_{n}"

            # bounding box
            width_bound= len(top_labels)+1
            title = "│" + self.simpleLieType.center(width_bound) + " │"
            border_top = "┌"+"─"*width_bound+"─┐"
            head_sep = "╞"+"═"*width_bound+"═╡"
            top_labels = "│" + top_labels + "  │"
            horiz = "│" + horiz.ljust(width_bound) + " │"
            border_bottom= "└"+"─"*width_bound+"─┘"


            lines.append(border_top)
            lines.append(title)
            lines.append(head_sep)
            lines.append(top_labels)
            lines.append(horiz)
            lines.append(border_bottom)

            print("\n".join(lines))
        elif self.simpleLieType[0]=='A':
            n = self.rank
            lines = []
            horiz = "   "
            if n > 7:
                mid_nodes = ["r_1 r_2", f"r_{n-3}", f"r_{n-2}", f"r_{n-1}"]
                latter_rules = [" ◯───◯─"+" ⋯ ─","─"*(len(mid_nodes[1])),"─"*(len(mid_nodes[2])),"─"*(len(mid_nodes[3])),""]
                horiz += "◯".join(latter_rules)
                top_labels = f"{' '*4}{mid_nodes[0]}{' '*3}{mid_nodes[1]} {mid_nodes[2]} {mid_nodes[3]}"
            else:
                horiz += "───".join("◯" for _ in range(n-1))
                top_labels = "   " + " ".join([f"r_{i+1}" for i in range(n-1)])
                horiz += "───◯"
            top_labels += " " + f"r_{n}"

            # bounding box
            width_bound= len(top_labels)+1
            title = "│" + self.simpleLieType.center(width_bound) + " │"
            border_top = "┌"+"─"*width_bound+"─┐"
            head_sep = "╞"+"═"*width_bound+"═╡"
            top_labels = "│" + top_labels + "  │"
            horiz = "│" + horiz.ljust(width_bound) + " │"
            border_bottom= "└"+"─"*width_bound+"─┘"


            lines.append(border_top)
            lines.append(title)
            lines.append(head_sep)
            lines.append(top_labels)
            lines.append(horiz)
            lines.append(border_bottom)

            print("\n".join(lines))

    def parabolic_grading(self, roots=None):
        if roots is None:
            roots=[]
        if isinstance(roots,int):
            roots=[roots]
        elif not isinstance(roots,(list,tuple)):
            raise TypeError(f'The `roots` parameter in `simpleLieAlgebra.parabolic_grading(roots)` should be either `None`, an `int`, or a list of integers in the range (1,...,{self.rank}) representing indices of simple roots as enumerated in the algebras Dynkin diagram (see `simpleLieAlgebra.root_space_summary()` for a summary of this indexing).') from None
        gradingVector = [sum([self.grading[idx-1][j] for idx in roots]) for j in range(self.dimension)]
        denom = 1
        for weight in gradingVector:
            if isinstance(weight,sp.Rational):
                if denom<weight.denominator:
                    denom=weight.denominator
        if denom>1:
            gradingVector=[denom*weight for weight in gradingVector]
        return gradingVector

    def parabolic_subalgebra(self,roots=None,label=None,basis_labels=None,register_in_vmf=None, return_Alg=False,use_non_positive_weights=False, format_as_subalgebra_class=False):
        if roots is None:
            roots=[]
        if isinstance(roots,int):
            roots=[roots]
        elif not isinstance(roots,(list,tuple)):
            raise TypeError(f'The `roots` parameter in `simpleLieAlgebra.parabolic_subalgebra(roots)` should be either `None`, an `int`, or a list of integers in the range (1,...,{self.rank}) representing indices of simple roots as enumerated in the algebras Dynkin diagram (see `simpleLieAlgebra.root_space_summary()` for a summary of this indexing).') from None
        newGrading = [sum([self.grading[idx-1][j] for idx in roots]) for j in range(self.dimension)]
        if format_as_subalgebra_class is True:
            parabolic = []
        subIndices = []
        filtered_grading = []
        if not isinstance(use_non_positive_weights,bool):
            use_non_positive_weights = False
        for count, weight in enumerate(newGrading):
            if (weight>=0 and use_non_positive_weights is False) or (weight<=0 and use_non_positive_weights is True):
                if format_as_subalgebra_class is True:
                    parabolic.append(self.basis[count])
                subIndices.append(count)
                filtered_grading.append(weight)
        denom = 1
        for weight in filtered_grading:
            if isinstance(weight,sp.Rational):
                if denom<weight.denominator:
                    denom=weight.denominator
        if denom>1:
            filtered_grading=[denom*weight for weight in filtered_grading]

        def truncateBySubInd(li):
            return [li[j] for j in subIndices]
        structureData = truncateBySubInd(self._structureData)
        structureData = [truncateBySubInd(plane) for plane in structureData]
        structureData = [[truncateBySubInd(li) for li in plane] for plane in structureData]
        if format_as_subalgebra_class is True:
            ignoredList = []
            if label is not None:
                ignoredList.append('label')
            if basis_labels is not None:
                ignoredList.append('basis_labels')
            if register_in_vmf is True:
                ignoredList.append('register_in_vmf')
            if len(ignoredList)==1:
                warnings.warn(f'A parameter value was supplied for `{ignoredList[0]}`, but `format_as_subalgebra_class=True` was set. The `subalgebra_class` is not tracked in the vmf, so this parameter value was ignored. A subalgebra_class instance was returned instead.')
            elif len(ignoredList)==2:
                warnings.warn(f'Parameter values were supplied for `{ignoredList[0]}` and `{ignoredList[1]}`, but `format_as_subalgebra_class=True` was set. The `subalgebra_class` is not tracked in the vmf, so these parameter values were ignored. A subalgebra_class instance was returned instead.')
            elif len(ignoredList)==3:
                warnings.warn(f'Parameter values were supplied for `{ignoredList[0]}`, `{ignoredList[1]}`, and `{ignoredList[2]}`, but `format_as_subalgebra_class=True` was set. The `subalgebra_class` is not tracked in the vmf, so these parameter values were ignored. `A subalgebra_class instance was returned instead.`')
            return subalgebra_class(parabolic, self, grading=[filtered_grading], _compressed_structure_data=structureData, _internal_lock=retrieve_passkey())
        if isinstance(label,str) or isinstance(basis_labels,(list,tuple,str)):
            register_in_vmf=True
        if register_in_vmf is True:
            if label is None:
                label = self.label+'_parabolic'
            if basis_labels is None:
                basis_labels = label
            elif (isinstance(basis_labels,(list,tuple)) and not all(isinstance(elem,str) for elem in basis_labels)) or not isinstance(basis_labels,str):
                raise TypeError('If supplying the optional parameter `basis_labels` to `simpleLieAlgebra.parabolic_subalgebra` then it should be either a string or list of strings') from None
            createAlgebra(structureData,label=label,basis_labels=basis_labels,grading=filtered_grading)
            if return_Alg is True:
                return _cached_caller_globals[label]
        if return_Alg is True:
            return algebra_class(structureData,grading=[filtered_grading])
        elif register_in_vmf is not True:
            warnings.warn('Optional keywords for the `parabolic_subalgebra` method indicate that nothing should be return returned or registered in the vmf. Probably that is not intended, in which case at least one keyword `label`, `basis_labels`, `register_in_vmf`, `return_Alg`, or `format_as_subalgebra_class` should be set differently.')


def createSimpleLieAlgebra(
    series: str,
    label: str = None,
    basis_labels: list = None,
    build_standard_mat_rep = False
):
    """
    Creates a simple (with 2 exceptions) complex Lie algebra specified from the classical
    series
        - A_n = sl(n+1)     for n>0
        - B_n = so(2n+1)    for n>0
        - C_n = sp(2n)      for n>0
        - D_n = so(2n)      for n>0 (not simple for n=1,2)


    Parameters
    ----------
    series : str
        The type and rank of the Lie algebra, e.g., "A1", "A2", ..., "Dn".
    label : str, optional
        Custom label for the Lie algebra. If not provided, defaults to a standard notation,
        like sl2 for A2 etc.
    basis_labels : list, optional
        Custom labels for the basis elements. If not provided, default labels will be generated.

    Returns
    -------
    algebra
        The resulting Lie algebra as an algebra instance.

    Raises
    ------
    ValueError
        If the series label is not recognized or not implemented.

    Notes
    -----
    - Currently supports only the A,B, and D series (special linear Lie algebras: A_n = sl(n+1), etc.).
    """
    # Extract series type and rank
    try:
        series_type, rank = series[0], int(series[1:])
        series_type = ''.join(c.upper() if c.islower() else c for c in series_type)
    except (IndexError, ValueError):
        raise ValueError(f"Invalid series format: {series}. Expected a letter 'A', 'B', 'C', 'D', 'E', 'F', or 'G' followed by a positive integer, like 'A1', 'B5', etc.") from None
    if rank <= 0:
            raise ValueError(f"Sequence index must be a positive integer, but got: {rank}.") from None

    def _generate_A_series_structure_data(n):
        matrix_dim = n+1

        # Basis elements
        hBasis = {'elems':dict(),'grading':dict()}
        offDiag = {'elems':dict(),'grading':dict()}

        repMatrix = [[0] * matrix_dim for _ in range(matrix_dim)]
        def elemWeights(idx1,idx2):
            wVec = []
            for idx in range(n):
                if idx1==idx:
                    if idx2==idx+1:
                        wVec.append(2)
                    else:
                        wVec.append(1)
                elif idx1==idx+1:
                    if idx2==idx:
                        wVec.append(-2)
                    else:
                        wVec.append(-1)
                elif idx2==idx:
                    wVec.append(-1)
                elif idx2==idx+1:
                    wVec.append(1)
                else:
                    wVec.append(0)
            return wVec

        for j in range(n+1):
            for k in range(j,n+1):
                # Diagonal (Cartan) element
                if j == k and j<n:
                    M = [row[:] for row in repMatrix]
                    for idx in range(n+1):
                        if idx==j+1:
                            M[idx][idx] = -1
                        elif idx==j:
                            M[idx][idx] = 1
                    hBasis['elems'][(j,k,0)] = M
                    hBasis['grading'][(j,k,0)] = [0]*n
                elif j!=k:
                    # off diagonal generators
                    MPlus = [row[:] for row in repMatrix]
                    MMinus = [row[:] for row in repMatrix]
                    MPlus[j][k]=1
                    MMinus[k][j]=1
                    offDiag['elems'][(j,k,1)] = MPlus
                    offDiag['grading'][(j,k,1)] = elemWeights(j,k)
                    offDiag['elems'][(k,j,1)] = MMinus
                    offDiag['grading'][(k,j,1)] = elemWeights(k,j)

        indexingKey = dict(enumerate(list(hBasis['grading'].keys())+list(offDiag['grading'].keys())))
        indexingKeyRev = {j:k for k,j in indexingKey.items()}
        LADimension = len(indexingKey)
        def _structureCoeffs(idx1,idx2):
            coeffs = [0]*LADimension
            if idx2==idx1:
                return coeffs
            if idx2<idx1:
                reSign = -1  
                idx2,idx1 = idx1,idx2
            else: 
                reSign = 1
            p10,p11,p12=indexingKey[idx1]
            p20,p21,p22=indexingKey[idx2]
            if p12==0:
                if p22==1:
                    coeffs[idx2]+=reSign*(int(p10==p20)-int(p10==p21)+int(p10+1==p21)-int(p10+1==p20))
            elif p12==1:
                if p22==1:
                    if p11==p20:
                        if p10==p21:
                            if p10<p11:
                                for idx in range(p10,p11):
                                    coeffs[indexingKeyRev[(idx,idx,0)]]=reSign
                            else:
                                for idx in range(p11,p10):
                                    coeffs[indexingKeyRev[(idx,idx,0)]]=-reSign
                        else:
                            coeffs[indexingKeyRev[(p10,p21,1)]]=reSign
                    elif p10==p21:
                        coeffs[indexingKeyRev[(p20,p11,1)]]=-reSign
            return coeffs


        _structure_data = [[_structureCoeffs(k,j) for j in range(LADimension)] for k in range(LADimension)]
        CartanSubalg=list(hBasis['elems'].values())
        matrixBasis=CartanSubalg+list(offDiag['elems'].values())
        gradingVecs=list(hBasis['grading'].values())+list(offDiag['grading'].values())
        return _structure_data, list(zip(*gradingVecs)), CartanSubalg, matrixBasis

    def _generate_B_series_structure_data(n):
        matrix_dim = 2*n+1

        # Basis elements
        hBasis = {'elems':dict(),'grading':dict()}
        GPlus = {'elems':dict(),'grading':dict()}
        GMinus = {'elems':dict(),'grading':dict()}
        DPlus = {'elems':dict(),'grading':dict()}
        DMinus = {'elems':dict(),'grading':dict()}

        skew_symmetric = [[0] * matrix_dim for _ in range(matrix_dim)]
        def gPlusWeights(idx1,idx2):
            wVec = []
            for idx in range(n-1):
                if (idx1<=idx and idx2<=idx) or (idx1>idx and idx2>idx):
                    wVec.append(0)
                elif idx1<=idx:
                    wVec.append(1)
                else:
                    wVec.append(-1)
            wVec.append(0)
            return wVec
        def gMinusWeights(idx1,idx2):
            wVec = []
            sign = 1 if idx2<idx1 else -1
            for idx in range(n-1):
                if (idx1<=idx and idx2<=idx):
                    wVec.append(sign)
                elif (idx1>idx and idx2>idx):
                    wVec.append(-sign)
                else:
                    wVec.append(0)
            wVec.append(sign)
            return wVec
        def DWeights(idx1,sign):
            wVec = []
            for idx in range(n-1):
                if (idx1<=idx):
                    wVec.append(-sign*sp.Rational(1,2))
                else:
                    wVec.append(-sign*sp.Rational(-1,2))
            wVec.append(-sign*sp.Rational(1,2))
            return wVec

        for j, k in carProd(range(n), range(n)):
            # Diagonal (Cartan) element
            if j == k and j<n-1:
                M = [row[:] for row in skew_symmetric]
                for idx in range(n):
                    if idx>j:
                        M[2*idx][2*idx+1] = -sp.I/2
                        M[2*idx+1][2*idx] = sp.I/2
                    else:
                        M[2*idx][2*idx+1] = sp.I/2
                        M[2*idx+1][2*idx] = -sp.I/2
                hBasis['elems'][(j,k,0)] = M
                hBasis['grading'][(j,k,0)] = [0]*n
                if j+2==n:
                    M = [row[:] for row in skew_symmetric]
                    for idx in range(n):
                        M[2*idx][2*idx+1] = sp.I/2
                        M[2*idx+1][2*idx] = -sp.I/2
                    hBasis['elems'][(j+1,k+1,0)] = M
                    hBasis['grading'][(j+1,k+1,0)] = [0]*n
            elif j!=k:
                # “+” generator
                MPlus = [row[:] for row in skew_symmetric]
                MPlus[2*j][2*k]     =  1
                MPlus[2*k][2*j]     = -1
                MPlus[2*j+1][2*k+1] =  1
                MPlus[2*k+1][2*j+1] = -1
                MPlus[2*j][2*k+1]   =  sp.I
                MPlus[2*k+1][2*j]   = -sp.I
                MPlus[2*j+1][2*k]   = -sp.I
                MPlus[2*k][2*j+1]   =  sp.I
                GPlus['elems'][(j,k,1)] = MPlus
                GPlus['grading'][(j,k,1)] = gPlusWeights(j,k)

                # “–” generator
                if j<k:
                    MMinus = [row[:] for row in skew_symmetric]
                    MMinus[2*j][2*k]     =  1
                    MMinus[2*k][2*j]     = -1
                    MMinus[2*j+1][2*k+1] = -1
                    MMinus[2*k+1][2*j+1] =  1
                    MMinus[2*j][2*k+1]   =  sp.I
                    MMinus[2*k+1][2*j]   = -sp.I
                    MMinus[2*j+1][2*k]   =  sp.I
                    MMinus[2*k][2*j+1]   = -sp.I
                    GMinus['elems'][(j,k,-1)] = MMinus
                    GMinus['grading'][(j,k,-1)] = gMinusWeights(j,k)
                else: # k<j
                    MMinus = [row[:] for row in skew_symmetric]
                    MMinus[2*k][2*j]     =  1
                    MMinus[2*j][2*k]     = -1
                    MMinus[2*k+1][2*j+1] = -1
                    MMinus[2*j+1][2*k+1] =  1
                    MMinus[2*k][2*j+1]   = -sp.I
                    MMinus[2*j+1][2*k]   = sp.I
                    MMinus[2*k+1][2*j]   = -sp.I
                    MMinus[2*j][2*k+1]   = sp.I
                    GMinus['elems'][(j,k,-1)] = MMinus
                    GMinus['grading'][(j,k,-1)] = gMinusWeights(j,k)
        for j in range(n):
            MPlus = [row[:] for row in skew_symmetric]
            MMinus = [row[:] for row in skew_symmetric]
            MPlus[2*j][2*n] = 1
            MPlus[2*n][2*j] = -1
            MPlus[2*j+1][2*n] = sp.I
            MPlus[2*n][2*j+1] = -sp.I
            MMinus[2*j][2*n] = 1
            MMinus[2*n][2*j] = -1
            MMinus[2*j+1][2*n] = -sp.I
            MMinus[2*n][2*j+1] = sp.I
            DPlus['elems'][(j,2*n,2)] = MPlus
            DPlus['grading'][(j,2*n,2)] = DWeights(j,1)
            DMinus['elems'][(j,2*n,-2)] = MMinus
            DMinus['grading'][(j,2*n,-2)] = DWeights(j,-1)

        indexingKey = dict(enumerate(list(hBasis['grading'].keys())+list(GPlus['grading'].keys())+list(GMinus['grading'].keys())+list(DPlus['grading'].keys())+list(DMinus['grading'].keys())))
        indexingKeyRev = {j:k for k,j in indexingKey.items()}
        LADimension = len(indexingKey)
        CSDict = {idx:{0:1,n-1:1} if idx==0 else {idx:1,idx-1:-1} for idx in range(n)} # Cartan subalgebra basis transform indexing
        CSDictInv = {idx:{j:-1 if j>idx else 1 for j in range(n)} for idx in range(n-1)}|{n-1:{j:1 for j in range(n)}}
        # _cached_caller_globals['debug']={'indexingKey':indexingKey}
        def _structureCoeffs(idx1,idx2):
            coeffs = [0]*LADimension
            if idx2==idx1:
                return coeffs
            if idx2<idx1:
                reSign = -1  
                idx2,idx1 = idx1,idx2
            else: 
                reSign = 1
            idx1,idx2 = sorted([idx1,idx2])
            p10,p11,p12=indexingKey[idx1]
            p20,p21,p22=indexingKey[idx2]
            if p12==0:
                for term, scale in CSDictInv[p10].items():
                    if p22==1:
                        coeffs[idx2]+=scale*reSign*(int(term==p20)-int(term==p21))*sp.Rational(1,2)
                    elif p22==-1:
                        sign = -reSign if p20<p21 else reSign
                        coeffs[idx2] += scale*sign*(int(term==p20)+int(term==p21))*sp.Rational(1,2)
                    elif p22==2:
                        coeffs[idx2] += -scale*(int(term==p20))*sp.Rational(1,2)*reSign
                    elif p22==-2:
                        coeffs[idx2] += scale*(int(term==p20))*sp.Rational(1,2)*reSign
            elif p12==1:
                if p22==1:
                    if p11==p20:
                        if p10==p21:
                            #l(p10)-l(p11)
                            for t, s in CSDict[p10].items():
                                coeffs[t]+=reSign*4*s
                            for t, s in CSDict[p11].items():
                                coeffs[t]+=-reSign*4*s
                        else:
                            coeffs[indexingKeyRev[(p10,p21,1)]]+=2*reSign
                    elif p10==p21:
                        coeffs[indexingKeyRev[(p20,p11,1)]]+=-2*reSign
                elif p22==-1:
                    slope1 = 1 if p10<p11 else -1
                    slope2 = 1 if p20<p21 else -1
                    if p10==p20:
                        if not(slope1==-1 and slope2==-1):
                            if p11<p21:
                                coeffs[indexingKeyRev[(p11,p21,-1)]]+=-2*reSign
                            elif p21<p11:
                                if not(slope1==1 and slope2==-1):
                                    coeffs[indexingKeyRev[(p21,p11,-1)]]+=2*reSign
                    elif p11==p21:
                        if not(slope1==1 and slope2==1):
                            if p10<p20:
                                coeffs[indexingKeyRev[(p20,p10,-1)]]+=2*reSign
                            elif p20<p10:
                                if not(slope1==-1 and slope2==1):
                                    coeffs[indexingKeyRev[(p10,p20,-1)]]+=-2*reSign
                    elif p11==p20:
                        if not(slope1==1 and slope2==1) and not(slope1==-1 and slope2==1):
                            if p10<p21:
                                coeffs[indexingKeyRev[(p21,p10,-1)]]=-2*reSign
                            elif p21<p10:
                                coeffs[indexingKeyRev[(p10,p21,-1)]]=2*reSign
                    elif p10==p21:
                        if not(slope1==-1 and slope2==-1) and not(slope1==1 and slope2==-1):
                            if p11<p20:
                                coeffs[indexingKeyRev[(p11,p20,-1)]]=2*reSign
                            elif p20<p11:
                                coeffs[indexingKeyRev[(p20,p11,-1)]]=-2*reSign
                elif p22==2:
                    if p10==p20:
                        coeffs[indexingKeyRev[(p11,p21,2)]]=-2*reSign
                elif p22==-2:
                    if p11==p20:
                        coeffs[indexingKeyRev[(p10,p21,-2)]]=2*reSign
            elif p12==-1:
                slope1 = 1 if p10<p11 else -1
                slope2 = 1 if p20<p21 else -1
                if p22==-1:
                    sign2 = 1 if p10<p11 else -1
                    if (p10<p11 and p20<p21) or (p10>p11 and p20>p21):
                        pass
                    elif p11==p20:
                        if p10==p21:
                            # plus/minus (l(p10)+l(p11))
                            for t, s in CSDict[p10].items():
                                coeffs[t]+=sign2*reSign*4*s
                            for t, s in CSDict[p11].items():
                                coeffs[t]+=sign2*reSign*4*s
                        else:
                            if sign2==1:
                                coeffs[indexingKeyRev[(p21,p10,1)]]+=2*reSign*sign2
                            else:
                                coeffs[indexingKeyRev[(p10,p21,1)]]+=2*reSign*sign2
                    elif p10==p21:
                        if sign2==1:
                            coeffs[indexingKeyRev[(p20,p11,1)]]+=2*reSign*sign2
                        else:
                            coeffs[indexingKeyRev[(p11,p20,1)]]+=2*reSign*sign2
                    elif p10==p20 and p21!=p11:
                        if sign2==1:
                            coeffs[indexingKeyRev[(p21,p11,1)]]+=-2*reSign*sign2
                        else:
                            coeffs[indexingKeyRev[(p11,p21,1)]]+=-2*reSign*sign2
                    elif p11==p21 and p10!=p20:
                        if sign2==1:
                            coeffs[indexingKeyRev[(p20,p10,1)]]+=-2*reSign*sign2
                        else:
                            coeffs[indexingKeyRev[(p10,p20,1)]]+=-2*reSign*sign2
                elif p22==2:
                    if slope1==-1:
                        if p11==p20:
                            coeffs[indexingKeyRev[(p10,p21,-2)]]=-2*reSign
                        elif p10==p20:
                            coeffs[indexingKeyRev[(p11,p21,-2)]]=2*reSign
                elif p22==-2:
                    if p10==p20 and slope1==1:
                        coeffs[indexingKeyRev[(p11,p21,2)]]=-2*reSign
                    if p11==p20:
                        if slope1==1:
                            coeffs[indexingKeyRev[(p10,p21,2)]]=2*reSign
            elif p12==2:
                if p22==2:
                    if p10<p20:
                        coeffs[indexingKeyRev[(p10,p20,-1)]]=-reSign
                    elif p20<p10:
                        coeffs[indexingKeyRev[(p20,p10,-1)]]=reSign
                if p22==-2:
                    if p10==p20:
                        for term, scale in CSDict[p10].items():
                            coeffs[term]=2*scale*reSign
                    elif p10<p20:
                        coeffs[indexingKeyRev[(p20,p10,1)]]=reSign
                    else:
                        coeffs[indexingKeyRev[(p20,p10,1)]]=reSign
            elif p12==-2:
                if p22==-2:
                    if p10<p20:
                        coeffs[indexingKeyRev[(p20,p10,-1)]]=-reSign
                    elif p20<p10:
                        coeffs[indexingKeyRev[(p10,p20,-1)]]=reSign
            return coeffs


        _structure_data = [[_structureCoeffs(k,j) for j in range(LADimension)] for k in range(LADimension)]
        CartanSubalg=list(hBasis['elems'].values())
        matrixBasis=CartanSubalg+list(GPlus['elems'].values())+list(GMinus['elems'].values())+list(DPlus['elems'].values())+list(DMinus['elems'].values())
        gradingVecs=list(hBasis['grading'].values())+list(GPlus['grading'].values())+list(GMinus['grading'].values())+list(DPlus['grading'].values())+list(DMinus['grading'].values())
        return _structure_data, list(zip(*gradingVecs)), CartanSubalg, matrixBasis

    def generate_C_series_structure_data(n):
        """
        Generates the structure data and weight vectors for the C_n series 
        (symplectic Lie algebra: sp(2n)).

        Parameters
        ----------
        n : int
            The rank of the C-series Lie algebra (2n is the matrix dimension).

        Returns
        -------
        tuple
            - basis (list): A 3-dimensional list representing the structure data for sp(2n).
                            Each element is a 2D list of lists representing a matrix.
            - weight_vectors (list): A single weight vector for the basis elements, 
                                    representing a grading of the algebra.

        Notes
        -----
        - Basis matrices are partitioned into nxn blocks and constructed in three groups:
        1. Lower block triangular: [[0, 0], [S_{j,k}, 0]] (weight = -1).
        2. Block diagonal: [[E_{j,k}, 0], [0, -E_{k,j}]] (weight = 0).
        3. Upper block triangular: [[0, S_{j,k}], [0, 0]] (weight = 1).
        """
        # Dimension of the full matrices
        matrix_dim = 2 * n

        # Basis elements and weight vector
        basis = []
        weight_vector = []

        # Step 1: Create symmetric matrices S_{j,k} = E_{j,k} + E_{k,j} for j ≤ k
        symmetric_matrices = []
        for j in range(n):
            for k in range(j, n):
                S = [[0] * n for _ in range(n)]
                S[j][k] = 1
                if j != k:
                    S[k][j] = 1
                symmetric_matrices.append(S)

        # Step 2: Create pairs P_{j,k} = (E_{j,k}, -E_{k,j})
        matrix_pairs = []
        for j in range(n):
            for k in range(n):
                P1 = [[0] * n for _ in range(n)]
                P2 = [[0] * n for _ in range(n)]
                P1[j][k] = 1
                P2[k][j] = -1
                matrix_pairs.append((P1, P2))

        # Step 3: Create basis matrices in three groups
        # Group 1: Lower block triangular [[0, 0], [S_{j,k}, 0]] (weight = -1)
        for S in symmetric_matrices:
            lower_triangular = [[0] * matrix_dim for _ in range(matrix_dim)]
            # Insert S into the lower-left block
            for i in range(n):
                for j in range(n):
                    lower_triangular[n + i][j] = S[i][j]
            basis.append(lower_triangular)
            weight_vector.append(-1)

        # Group 2: Block diagonal [[E_{j,k}, 0], [0, -E_{k,j}]] (weight = 0)
        for P1, P2 in matrix_pairs:
            block_diagonal = [[0] * matrix_dim for _ in range(matrix_dim)]
            # Insert P1 into the top-left block and P2 into the bottom-right block
            for i in range(n):
                for j in range(n):
                    block_diagonal[i][j] = P1[i][j]
                    block_diagonal[n + i][n + j] = P2[i][j]
            basis.append(block_diagonal)
            weight_vector.append(0)

        # Group 3: Upper block triangular [[0, S_{j,k}], [0, 0]] (weight = 1)
        for S in symmetric_matrices:
            upper_triangular = [[0] * matrix_dim for _ in range(matrix_dim)]
            # Insert S into the upper-right block
            for i in range(n):
                for j in range(n):
                    upper_triangular[i][n + j] = S[i][j]
            basis.append(upper_triangular)
            weight_vector.append(1)

        # Return basis and weight vector wrapped in a list (to allow multiple gradings)
        return basis, [weight_vector]

    def _generate_D_series_structure_data(n):
        matrix_dim = 2 * n

        # Basis elements
        hBasis = {'elems':dict(),'grading':dict()}
        GPlus = {'elems':dict(),'grading':dict()}
        GMinus = {'elems':dict(),'grading':dict()}

        skew_symmetric = [[0] * matrix_dim for _ in range(matrix_dim)]
        def gPlusWeights(idx1,idx2):
            wVec = []
            for idx in range(n-1):
                if (idx1<=idx and idx2<=idx) or (idx1>idx and idx2>idx):
                    wVec.append(0)
                elif idx1<=idx:
                    wVec.append(-1)
                else:
                    wVec.append(1)
            wVec.append(0)
            return wVec
        def gMinusWeights(idx1,idx2):
            wVec = []
            sign = -1 if idx2<idx1 else 1
            for idx in range(n-1):
                if (idx1<=idx and idx2<=idx):
                    wVec.append(sign)
                elif (idx1>idx and idx2>idx):
                    wVec.append(-sign)
                else:
                    wVec.append(0)
            wVec.append(sign)
            return wVec

        for j, k in carProd(range(n), range(n)):
            # Diagonal (Cartan) element
            if j == k and j<n-1:
                M = [row[:] for row in skew_symmetric]
                for idx in range(n):
                    if idx>j:
                        M[2*idx][2*idx+1] = -sp.I/2
                        M[2*idx+1][2*idx] = sp.I/2
                    else:
                        M[2*idx][2*idx+1] = sp.I/2
                        M[2*idx+1][2*idx] = -sp.I/2
                hBasis['elems'][(j,k,0)] = M
                hBasis['grading'][(j,k,0)] = [0]*n
                if j+2==n:
                    M = [row[:] for row in skew_symmetric]
                    for idx in range(n):
                        M[2*idx][2*idx+1] = sp.I/2
                        M[2*idx+1][2*idx] = -sp.I/2
                    hBasis['elems'][(j+1,k+1,0)] = M
                    hBasis['grading'][(j+1,k+1,0)] = [0]*n
            elif j!=k:
                # “+” generator
                MPlus = [row[:] for row in skew_symmetric]
                MPlus[2*j][2*k]     =  1
                MPlus[2*k][2*j]     = -1
                MPlus[2*j+1][2*k+1] =  1
                MPlus[2*k+1][2*j+1] = -1
                MPlus[2*j][2*k+1]   =  sp.I
                MPlus[2*k+1][2*j]   = -sp.I
                MPlus[2*j+1][2*k]   = -sp.I
                MPlus[2*k][2*j+1]   =  sp.I
                GPlus['elems'][(j,k,1)] = MPlus
                GPlus['grading'][(j,k,1)] = gPlusWeights(j,k)

                # “–” generator
                if j<k:
                    MMinus = [row[:] for row in skew_symmetric]
                    MMinus[2*j][2*k]     =  1
                    MMinus[2*k][2*j]     = -1
                    MMinus[2*j+1][2*k+1] = -1
                    MMinus[2*k+1][2*j+1] =  1
                    MMinus[2*j][2*k+1]   =  sp.I
                    MMinus[2*k+1][2*j]   = -sp.I
                    MMinus[2*j+1][2*k]   =  sp.I
                    MMinus[2*k][2*j+1]   = -sp.I
                    GMinus['elems'][(j,k,-1)] = MMinus
                    GMinus['grading'][(j,k,-1)] = gMinusWeights(j,k)
                else: # k<j
                    MMinus = [row[:] for row in skew_symmetric]
                    MMinus[2*k][2*j]     =  1
                    MMinus[2*j][2*k]     = -1
                    MMinus[2*k+1][2*j+1] = -1
                    MMinus[2*j+1][2*k+1] =  1
                    MMinus[2*k][2*j+1]   = -sp.I
                    MMinus[2*j+1][2*k]   = sp.I
                    MMinus[2*k+1][2*j]   = -sp.I
                    MMinus[2*j][2*k+1]   = sp.I
                    GMinus['elems'][(j,k,-1)] = MMinus
                    GMinus['grading'][(j,k,-1)] = gMinusWeights(j,k)

        indexingKey = dict(enumerate(list(hBasis['grading'].keys())+list(GPlus['grading'].keys())+list(GMinus['grading'].keys())))
        indexingKeyRev = {j:k for k,j in indexingKey.items()}
        LADimension = len(indexingKey)
        CSDict = {idx:{0:1,n-1:1} if idx==0 else {idx:1,idx-1:-1} for idx in range(n)} # Cartan subalgebra basis transform indexing
        CSDictInv = {idx:{j:-1 if j>idx else 1 for j in range(n)} for idx in range(n-1)}|{n-1:{j:1 for j in range(n)}}
        def _structureCoeffs(idx1,idx2):
            coeffs = [0]*LADimension
            if idx2==idx1:
                return coeffs
            if idx2<idx1:
                reSign = -1  
                idx2,idx1 = idx1,idx2
            else: 
                reSign = 1
            idx1,idx2 = sorted([idx1,idx2])
            p10,p11,p12=indexingKey[idx1]
            p20,p21,p22=indexingKey[idx2]
            if p12==0:
                for term, scale in CSDictInv[p10].items():
                    if p22==1:
                        coeffs[idx2]+=scale*reSign*(int(term==p20)-int(term==p21))*sp.Rational(1,2)
                    elif p22==-1:
                        sign = -reSign if p20<p21 else reSign
                        coeffs[idx2] += scale*sign*(int(term==p20)+int(term==p21))*sp.Rational(1,2)
            elif p12==1:
                if p22==1:
                    if p11==p20:
                        if p10==p21:
                            #l(p10)-l(p11)
                            for t, s in CSDict[p10].items():
                                coeffs[t]+=reSign*4*s
                            for t, s in CSDict[p11].items():
                                coeffs[t]+=-reSign*4*s
                        else:
                            coeffs[indexingKeyRev[(p10,p21,1)]]+=2*reSign
                    elif p10==p21:
                        coeffs[indexingKeyRev[(p20,p11,1)]]+=-2*reSign
                else:
                    slope1 = 1 if p10<p11 else -1
                    slope2 = 1 if p20<p21 else -1
                    if p10==p20:
                        if not(slope1==-1 and slope2==-1):
                            if p11<p21:
                                coeffs[indexingKeyRev[(p11,p21,-1)]]+=-2*reSign
                            elif p21<p11:
                                if not(slope1==1 and slope2==-1):
                                    coeffs[indexingKeyRev[(p21,p11,-1)]]+=2*reSign
                    elif p11==p21:
                        if not(slope1==1 and slope2==1):
                            if p10<p20:
                                coeffs[indexingKeyRev[(p20,p10,-1)]]+=2*reSign
                            elif p20<p10:
                                if not(slope1==-1 and slope2==1):
                                    coeffs[indexingKeyRev[(p10,p20,-1)]]+=-2*reSign
                    elif p11==p20:
                        if not(slope1==1 and slope2==1) and not(slope1==-1 and slope2==1):
                            if p10<p21:
                                coeffs[indexingKeyRev[(p21,p10,-1)]]=-2*reSign
                            elif p21<p10:
                                coeffs[indexingKeyRev[(p10,p21,-1)]]=2*reSign
                    elif p10==p21:
                        if not(slope1==-1 and slope2==-1) and not(slope1==1 and slope2==-1):
                            if p11<p20:
                                coeffs[indexingKeyRev[(p11,p20,-1)]]=2*reSign
                            elif p20<p11:
                                coeffs[indexingKeyRev[(p20,p11,-1)]]=-2*reSign
            else:
                sign2 = 1 if p10<p11 else -1
                if (p10<p11 and p20<p21) or (p10>p11 and p20>p21):
                    pass
                elif p11==p20:
                    if p10==p21:
                        # plus/minus (l(p10)+l(p11))
                        for t, s in CSDict[p10].items():
                            coeffs[t]+=sign2*reSign*4*s
                        for t, s in CSDict[p11].items():
                            coeffs[t]+=sign2*reSign*4*s
                    else:
                        if sign2==1:
                            coeffs[indexingKeyRev[(p21,p10,1)]]+=2*reSign*sign2
                        else:
                            coeffs[indexingKeyRev[(p10,p21,1)]]+=2*reSign*sign2
                elif p10==p21:
                    if sign2==1:
                        coeffs[indexingKeyRev[(p20,p11,1)]]+=2*reSign*sign2
                    else:
                        coeffs[indexingKeyRev[(p11,p20,1)]]+=2*reSign*sign2
                elif p10==p20 and p21!=p11:
                    if sign2==1:
                        coeffs[indexingKeyRev[(p21,p11,1)]]+=-2*reSign*sign2
                    else:
                        coeffs[indexingKeyRev[(p11,p21,1)]]+=-2*reSign*sign2
                elif p11==p21 and p10!=p20:
                    if sign2==1:
                        coeffs[indexingKeyRev[(p20,p10,1)]]+=-2*reSign*sign2
                    else:
                        coeffs[indexingKeyRev[(p10,p20,1)]]+=-2*reSign*sign2
            return coeffs


        _structure_data = [[_structureCoeffs(k,j) for j in range(LADimension)] for k in range(LADimension)]
        CartanSubalg=list(hBasis['elems'].values())
        matrixBasis=CartanSubalg+list(GPlus['elems'].values())+list(GMinus['elems'].values())
        gradingVecs=list(hBasis['grading'].values())+list(GPlus['grading'].values())+list(GMinus['grading'].values())
        return _structure_data, list(zip(*gradingVecs)), CartanSubalg, matrixBasis

    if series_type == "A":
        default_label = f"sl{rank + 1}" if label is None else label
        # Compute structure data for sl(n+1)
        structure_data, grading,CartanSubalgebra, matrixBasis = _generate_A_series_structure_data(rank)
        # Create and return the Lie algebra
        passkey = retrieve_passkey()
        if build_standard_mat_rep is True:
            return createAlgebra(matrixBasis,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                process_matrix_rep=True,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )
        else:
            return createAlgebra(structure_data,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )

    elif series_type == "B":
        default_label = f"so{2*rank + 1}" if label is None else label
        # Compute structure data for so(2n+1)
        structure_data, grading,CartanSubalgebra, matrixBasis = _generate_B_series_structure_data(rank)
        # Create and return the Lie algebra
        passkey = retrieve_passkey()
        if build_standard_mat_rep is True:
            return createAlgebra(matrixBasis,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                process_matrix_rep=True,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )
        else:
            return createAlgebra(structure_data,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )

    elif series_type == "C":
        raise ValueError("The C series (i.e., C_1=sp(2), C2=sp(4), ...) is not yet supported by `createSimpleLieAlgebra`. A future update will include support the C series.") from None

    elif series_type == "D":
        default_label = f"so{2*rank}" if label is None else label
        # Compute structure data for so(2n)
        structure_data, grading,CartanSubalgebra, matrixBasis = _generate_D_series_structure_data(rank)
        # Create and return the Lie algebra
        passkey = retrieve_passkey()
        if build_standard_mat_rep is True:
            return createAlgebra(matrixBasis,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                process_matrix_rep=True,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )
        else:
            return createAlgebra(structure_data,
                label=default_label,
                basis_labels=basis_labels,
                grading=grading,
                preferred_representation=matrixBasis,
                _simple={'lockKey':passkey,'CartanSubalgebra':CartanSubalgebra,'type':[series_type,rank]}
            )

    elif series_type+str(rank) in {'G2', 'F4', 'E6', 'E7','E8'}:
        raise ValueError("Exceptional Lie algebras are not yet supported by `createSimpleLieAlgebra`.") from None

    else:
        raise ValueError(f"Invalid series parameter format: {series}. Expected a letter 'A', 'B', 'C', 'D', 'E', 'F', or 'G' followed by a positive integer, like 'A1', 'B5', etc. For the exceptional LA labels 'E', 'F', and 'G' the integer must be among the classified types (i.e., only 'G2', 'F4', 'E6', 'E7', and 'E8' are admissible).") from None

def createFiniteAlg(
        obj,
        label,
        basis_labels=None,
        grading=None,
        format_sparse=False,
        process_matrix_rep=False,
        preferred_representation=None,
        verbose=False,
        assume_skew=False, 
        assume_Lie_alg=False,
        basis_order_for_supplied_str_eqns=None,
        _simple=None
    ):
    warnings.warn(
        "`createFiniteAlg` has been deprecated as it is being replaced with a more genera function. "
        "It will be removed in 2026. Use `createAlgebra` instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return createAlgebra(obj, label, basis_labels=basis_labels, grading=grading, format_sparse=format_sparse, process_matrix_rep=process_matrix_rep, preferred_representation=preferred_representation, verbose=verbose, assume_skew=assume_skew, assume_Lie_alg=assume_Lie_alg, basis_order_for_supplied_str_eqns=basis_order_for_supplied_str_eqns, _simple=_simple)
def createAlgebra(
    obj,
    label,
    basis_labels=None,
    grading=None,
    format_sparse=False,
    process_matrix_rep=False,
    preferred_representation=None,
    verbose=False,
    assume_skew=False, 
    assume_Lie_alg=False,
    basis_order_for_supplied_str_eqns=None,
    _simple=None
):
    """
    Registers an algebra object and its basis elements in the caller's global namespace,
    and adds them to the variable_registry for tracking in the Variable Management Framework.

    Parameters
    ----------
    obj : algebra, structure data, or list of algebra_element_class
        The algebra object (an instance of algebra), the structure data used to create one,
        or a list of algebra_element_class instances with the same parent algebra.
    label : str
        The label used to reference the algebra object in the global namespace.
    basis_labels : list, optional
        A list of custom labels for the basis elements of the algebra.
        If not provided, default labels will be generated.
    grading : list of lists or list, optional
        A list specifying the grading(s) of the algebra.
    format_sparse : bool, optional
        Whether to use sparse arrays when creating the algebra object.
    process_matrix_rep : bool, optional
        Whether to compute and store the matrix representation of the algebra.
    verbose : bool, optional
        If True, provides detailed feedback during the creation process.
    """

    if get_dgcv_category(obj)=='Tanaka_symbol':
        if grading is not None:
            warnings.warn('When processing a `Tanaka_symbol` object, `createAlgebra` uses the symbol\'s pre-defined grading rather than a manually supplied grading. You are getting this warning because an additional grading was manually supplied. To apply the custom grading instead, extract the symbol object\'s structure data using `Tanaka_symbol.export_algebra_data()`, and then pass that to `createAlgebra`.')
        symbolData = obj.export_algebra_data(_internal_call_lock=retrieve_passkey())
        if isinstance(symbolData,str):
            raise TypeError(symbolData+' So no `createAlgebra` did not instantiate a new algebra.') from None
        obj = symbolData['structure_data']
        grading = symbolData['grading']

    passkey = retrieve_passkey()
    if label in listVar(algebras_only=True) and get_dgcv_settings_registry()['forgo_warnings'] is not True:
        if isinstance(_simple,dict) and _simple.get('lockKey',None)==passkey:
            callFunction = 'createSimpleLieAlgebra'
        else:
            callFunction = 'createAlgebra'
        warnings.warn(f'`{callFunction}` was called with a `label` parameter already assigned to another algebra, so `{callFunction}` will overwrite the other algebra in the VMF and global namespace.')
        clearVar(label)

    def extract_structure_from_elements(elements):
        """
        Computes structure constants and validates linear independence from a list of algebra_element_class.

        Parameters
        ----------
        elements : list of algebra_element_class
            A list of algebra_element_class instances.

        Returns
        -------
        structure_data : list of lists of lists
            The structure constants for the subalgebra spanned by the elements.

        Raises
        ------
        ValueError
            If the elements are not linearly independent or not closed under the algebra product.
        """
        if not elements or not all(isinstance(el, algebra_element_class) for el in elements):
            raise ValueError(
                "Invalid input: All elements must be instances of algebra_element_class."
            ) from None

        # Check that all elements have the same parent algebra
        parent_algebra = elements[0].algebra
        if not all(el.algebra == parent_algebra for el in elements):
            raise ValueError(
                "All algebra_element_class instances must share the same parent algebra."
            ) from None

        try:
            # Use the parent algebra's is_subspace_subalgebra method
            result = parent_algebra.is_subspace_subalgebra(
                elements, return_structure_data=True
            )
        except ValueError as e:
            raise ValueError(
                "Error during subalgebra validation. "
                "The input list of algebra_element_class instances must be linearly independent and closed under the algebra product. "
                f"Original error: {e}"
            ) from e

        if not result["linearly_independent"]:
            raise ValueError(
                "The input elements are not linearly independent. "
            ) from None

        if not result["closed_under_product"]:
            raise ValueError(
                "The input elements are not closed under the algebra product. "
            ) from None

        # Return structure data
        return result["structure_data"]

    # Validate or create the algebra object
    if isinstance(obj, algebra_class):
        if verbose:
            print(f"Using existing algebra instance: {label}")
        structure_data = obj.structureData
        dimension = obj.dimension
    elif isinstance(obj, (list,tuple)) and all(isinstance(el, algebra_element_class) for el in obj):
        if verbose:
            print("Creating algebra from list of algebra_element_class instances.")
        structure_data = extract_structure_from_elements(obj)
        dimension = len(obj)
    else:
        if verbose:
            print("processing structure data...")
        structure_data = _validate_structure_data(
            obj, process_matrix_rep=process_matrix_rep, assume_skew=assume_skew, assume_Lie_alg=assume_Lie_alg, basis_order_for_supplied_str_eqns=basis_order_for_supplied_str_eqns
        )
        dimension = len(structure_data)

    # Create or validate basis labels
    if basis_labels is None:
        basis_labels = [validate_label(f"{label}_{i+1}") for i in range(dimension)]
    elif isinstance(basis_labels,str):
        basis_labels = [validate_label(f"{basis_labels}_{i+1}") for i in range(dimension)]
    else:
        validate_label_list(basis_labels)

    # Process grading
    if grading is None:
        grading = [tuple([0] * dimension)]
    elif isinstance(grading, (list, tuple)) and all(
        isinstance(w, (int, sp.Expr)) for w in grading
    ):
        # Single grading vector
        if len(grading) != dimension:
            raise ValueError(
                f"Grading vector length ({len(grading)}) must match the algebra dimension ({dimension})."
            ) from None
        grading = [tuple(grading)]
    elif isinstance(grading, list) and all(
        isinstance(vec, (list, tuple)) for vec in grading
    ):
        # List of grading vectors
        for vec in grading:
            if len(vec) != dimension:
                raise ValueError(
                    f"Grading vector length ({len(vec)}) must match the algebra dimension ({dimension})."
                ) from None
        grading = [tuple(vec) for vec in grading]
    else:
        raise ValueError("Grading must be a single vector or a list of vectors.") from None

    if isinstance(_simple,dict) and _simple.get('lockKey',None)==passkey:
        algebra_obj = simpleLieAlgebra(
            structure_data=structure_data,
            grading=grading,
            format_sparse=format_sparse,
            process_matrix_rep=process_matrix_rep,
            preferred_representation=preferred_representation,
            _label=label,
            _basis_labels=basis_labels,
            _calledFromCreator=passkey,
            _simple_data = _simple
        )
    else:
        algebra_obj = algebra_class(
            structure_data=structure_data,
            grading=grading,
            format_sparse=format_sparse,
            process_matrix_rep=process_matrix_rep,
            preferred_representation=preferred_representation,
            _label=label,
            _basis_labels=basis_labels,
            _calledFromCreator=passkey,
        )

    assert (
        algebra_obj.basis is not None
    ), "Algebra object basis elements must be initialized."

    _cached_caller_globals.update({label: algebra_obj})
    _cached_caller_globals.update(zip(basis_labels, algebra_obj.basis))

    variable_registry = get_variable_registry()
    variable_registry["finite_algebra_systems"][label] = {
        "family_type": "algebra",
        "family_names": tuple(basis_labels),
        "family_values": tuple(algebra_obj.basis),
        "dimension": dimension,
        "grading": grading,
        "basis_labels": basis_labels,
        "structure_data": structure_data,
    }
    variable_registry["_labels"][label] = {
        "path": ("finite_algebra_systems", label),
        "children": set(basis_labels)
    }

    if verbose:
        print(f"Algebra '{label}' registered successfully.")
        print(
            f"Dimension: {dimension}, Grading: {grading}, Basis Labels: {basis_labels}"
        )




from .algebras.algebras_core import algebra_class
from .tensors import vector_space_class


class generalLinearGroup(algebra_class):
    def __init__(self, vector_space, grading=None, format_sparse=False, process_matrix_rep=False, _label=None, _basis_labels=None, _calledFromCreator=None, _callLock=None, _print_warning=None, _child_print_warning=None, _exclude_from_VMF=None):
        if not isinstance(vector_space,(algebra_class,vector_space_class)):
            raise TypeError('The `vector_space` parameter in the `generalLinearGroup` initializer must of type `FAClass` or `vector_space_class`.')
        super().__init__(vector_space, grading, format_sparse, process_matrix_rep, _label, _basis_labels, _calledFromCreator, _callLock, _print_warning, _child_print_warning, _exclude_from_VMF)



import random

import sympy as sp

from .._config import _cached_caller_globals
from .._safeguards import retrieve_passkey, retrieve_public_key
from ..dgcv_core import VF_bracket, VFClass, addVF, allToReal, variableProcedure
from ..solvers import solve_dgcv
from ..vmf import clearVar, listVar


def _validate_structure_data(data, process_matrix_rep=False, assume_skew=False, assume_Lie_alg=False, basis_order_for_supplied_str_eqns = None):
    if process_matrix_rep:
        if all(
            isinstance(sp.Matrix(obj), sp.Matrix)
            and len(set(sp.Matrix(obj).shape[:2]))<2
            for obj in data
        ):
            return algebraDataFromMatRep(data)
        else:
            raise ValueError(
                f"matrix representation prcessing requires a list of square matrices. Recieved: {data}"
            )

    if isinstance(data,(list,tuple)) and all(isinstance(obj, VFClass) for obj in data):
        return algebraDataFromVF(data)
    try:
        if isinstance(data, dict):
            if all(isinstance(key,tuple) and len(key)==2 and all(isinstance(idx,sp.Symbol) for idx in key) for key in data):
                if basis_order_for_supplied_str_eqns is None:
                    basis_order_for_supplied_str_eqns = []
                if not isinstance(basis_order_for_supplied_str_eqns,(list,tuple)) or any(not isinstance(var,sp.Symbol) for var in basis_order_for_supplied_str_eqns):
                    raise ValueError('If initializing an algebra from structure equations and supplying the `basis_order_for_supplied_str_eqns` parameter, this parameter should be a list of the sympy.Symbols instances appearing in the supplied structure equations.')
                for var in set(sum([list(key) for key in data.keys()],[])):
                    if var not in basis_order_for_supplied_str_eqns:
                        raise ValueError('If initializing an algebra from structure equations and supplying the `basis_order_for_supplied_str_eqns` parameter, this parameter should be a list containing all sympy.Symbols instances appearing in the supplied structure equations.')
                ordered_BV = basis_order_for_supplied_str_eqns
                zeroing = {var:0 for var in ordered_BV}
                new_data = dict()
                for idx_pair, val in data.items():
                    v1,v2 = idx_pair
                    idx1 = ordered_BV.index(v1)
                    idx2 = ordered_BV.index(v2)
                    if hasattr(val,'subs') and val.subs(zeroing)==0:
                        coeffs = []
                        for var in ordered_BV:
                            coeffs.append(sp.simplify(val.subs({var:1}).subs(zeroing)))
                        new_data[(idx2,idx1)]=tuple(coeffs)
                    else:
                        print(val,zeroing)
                        raise ValueError('If initializing an algebra from structure equations, supplied structure equations should be a dictionary whose keys are tuples of variables (`sympy.Symbol` class instances) and whose value is a linear combination of variables representing the product of the elements in the key tuple.')
                data = new_data
            if all(isinstance(key,tuple) and len(key)==2 and all(isinstance(idx,int) and idx>=0 for idx in key) for key in data):
                provided_index_bound = max(sum([list(key) for key in data.keys()],[]))
            else:
                raise ValueError("Structure data must be have one of several formats: It can be a list/tuple with 3D shape of size (x, x, x). Or it can be a dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep.")
            if all(isinstance(val,(tuple,list)) for val in data.values()):
                base_dims = list(len(val) for val in data.values())
                if len(set(base_dims))!=1 or base_dims[0]<provided_index_bound+1:
                    raise ValueError("If initializing an algebra algebra with structure data from a dictionairy, its keys should be (i,j) index tuples and its values should be tuples of coefficients from the product of i and j basis elements. All values tuples must have the same length in particular. Indices in the keys must not exceed the length of value tuples - 1 (as indexing starts from 0!)")
                else:
                    base_dim = base_dims[0]
                if assume_skew or assume_Lie_alg:
                    seen = []
                    initial_keys = list(data.keys())
                    for idx in initial_keys:
                        if idx in seen:
                            pass
                        else:
                            invert_idx = (idx[1],idx[0])
                            if invert_idx in data.keys():
                                if any(j+k!=0 for j,k in zip(data[idx],data[invert_idx])):
                                    raise ValueError("Either `assume_skew=True` or `assume_Lie_alg=True` was passed to the algebra contructor, but the accompanying structure data was not skew symmetric.")
                            else:
                                data[invert_idx]=[-j for j in data[idx]]
                            seen+=[idx,invert_idx]
                data = [[list(data.get((j,k),[0]*base_dim)) for j in range(base_dim)] for k in range(base_dim)]
            else:
                raise ValueError("If initializing an algebra algebra with structure data from a dictionairy, its keys should be (i,j) index tuples and its values should be tuples of coefficients from the product of i and j basis elements. All values tuples must have the same length in particular.")

        # Check that the data is a 3D list-like structure
        if isinstance(data, (list,tuple)) and len(data) > 0 and isinstance(data[0], (list,tuple)):
            if len(data) == len(data[0]) == len(data[0][0]):
                return tuple(tuple(tuple(inner) for inner in outer) for outer in data)  # Return as a validated 3D list as tuples
            else:
                raise ValueError("Structure data must be a list with 3D shape of size (x, x, x). Or it can a  dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep.")
        else:
            raise ValueError("Structure data must be a list with 3D shape of size (x, x, x). Or it can a  dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep.")
    except Exception as e:
        raise ValueError(f"Invalid structure data format: {type(data)} - {e}")

def algebraDataFromVF(vector_fields):
    """
    Create the structure data array for a Lie algebra from a list of vector fields in *vector_fields*.

    This function computes the Lie algebra structure constants from a list of vector fields
    (instances of VFClass) defined on the same variable space. The returned structure data
    can be used to initialize an algebra instance.

    Parameters
    ----------
    vector_fields : list
        A list of VFClass instances, all defined on the same variable space with respect to the same basis.

    Returns
    -------
    list
        A 3D array-like list of lists of lists representing the Lie algebra structure data.

    Raises
    ------
    Exception
        If the vector fields do not span a Lie algebra or are not defined on a common basis.

    Notes
    -----
    This function dynamically chooses its approach to solve for the structure constants:
    - For smaller dimensional algebras, it substitutes pseudo-arbitrary values for the variables in `varSpaceLoc`
      based on a power function to create a system of linear equations.
    - For larger systems, where `len(varSpaceLoc)` raised to `len(vector_fields)` exceeds a threshold (default is 10,000),
      random rational numbers are used for substitution to avoid performance issues caused by large numbers.
    """
    # Define the product threshold for switching to random sampling
    product_threshold = 1

    # Check if all vector fields are defined on the same variable space
    if len(set([vf.varSpace for vf in vector_fields])) != 1:
        raise Exception(
            "algebraDataFromVF requires vector fields defined with respect to a common basis."
        )

    complexHandling = any(vf.dgcvType == "complex" for vf in vector_fields)
    if complexHandling:
        vector_fields = [allToReal(j) for j in vector_fields]
    varSpaceLoc = vector_fields[0].varSpace

    # Create temporary variables for solving structure constants
    tempVarLabel = "T" + retrieve_public_key()
    variableProcedure(tempVarLabel, len(vector_fields), _tempVar=retrieve_passkey())
    combiVFLoc = addVF(*[_cached_caller_globals[tempVarLabel][j] * vector_fields[j] for j in range(len(_cached_caller_globals[tempVarLabel]))])

    def computeBracket(j, k):
        """
        Compute and return the Lie bracket [vf_j, vf_k] and structure constants.

        Parameters
        ----------
        j : int
            Index of the first vector field.
        k : int
            Index of the second vector field.

        Returns
        -------
        list
            Structure constants for the Lie bracket of vf_j and vf_k.
        """
        if k <= j:
            return [0] * len(_cached_caller_globals[tempVarLabel])

        # Compute the Lie bracket
        bracket = VF_bracket(vector_fields[j], vector_fields[k]) - combiVFLoc

        if complexHandling:
            bracket = [allToReal(expr) for expr in bracket.coeffs]
        else:
            bracket = bracket.coeffs

        if len(varSpaceLoc) ** len(vector_fields) <= product_threshold:
            # Use the current system of pseudo-arbitrary substitutions
            bracketVals = list(
                set(
                    sum(
                        [
                            [
                                expr.subs(
                                    [
                                        (
                                            varSpaceLoc[i],
                                            sp.Rational((i + 1) ** sampling_index, 32),
                                        )
                                        for i in range(len(varSpaceLoc))
                                    ]
                                )
                                for expr in bracket
                            ]
                            for sampling_index in range(len(vector_fields))
                        ],
                        [],
                    )
                )
            )
        else:
            # Use random sampling system for larger cases
            def random_rational():
                return sp.Rational(random.randint(1, 1000), random.randint(1001, 2000))            
            bracketVals = list(
                set(
                    sum(
                        [
                            [
                                expr if not hasattr(expr,'subs') else
                                expr.subs(
                                    [
                                        (varSpaceLoc[i], random_rational())
                                        for i in range(len(varSpaceLoc))
                                    ]
                                )
                                for expr in bracket
                            ]
                            for _ in range(len(vector_fields))
                        ],
                        [],
                    )
                )
            )

        solutions = list(solve_dgcv(bracketVals, _cached_caller_globals[tempVarLabel]))
        if len(solutions) == 1:
            sol_values = solutions[0]
            substituted_constants = [
                expr.subs(sol_values)
                for expr in _cached_caller_globals[tempVarLabel]
            ]
            return substituted_constants
        else:
            raise Exception(
                f"Fields at positions {j} and {k} are not closed under Lie brackets."
            )

    # Precompute all necessary Lie brackets and store as 3D list
    structure_data = [[[0 for _ in vector_fields] for _ in vector_fields] for _ in vector_fields]

    for j in range(len(vector_fields)):
        for k in range(j + 1, len(vector_fields)):
            structure_data[j][k] = computeBracket(k, j)         # CHECK index order!!!
            structure_data[k][j] = [-elem for elem in structure_data[j][k]]

    # Clean up temporary variables
    clearVar(*listVar(temporary_only=True), report=False)

    return structure_data

def algebraDataFromMatRep(mat_list):
    """
    Create the structure data array for a Lie algebra from a list of matrices in *mat_list*.

    This function computes the Lie algebra structure constants from a matrix representation of a Lie algebra.
    The returned structure data can be used to initialize an algebra instance.

    Parameters
    ----------
    mat_list : list
        A list of square matrices of the same size representing the Lie algebra.

    Returns
    -------
    list
        A 3D list of lists of lists representing the Lie algebra structure data.

    Raises
    ------
    Exception
        If the matrices do not span a Lie algebra, or if the matrices are not square and of the same size.
    """
    if isinstance(mat_list, list):
        mListLoc = [sp.Matrix(j) for j in mat_list]
        shapeLoc = mListLoc[0].shape[0]
        indexRangeCap=len(mat_list)

        # Check that all matrices are square and of the same size
        if all(j.shape == (shapeLoc, shapeLoc) for j in mListLoc):
            tempVarLabel = "T" + retrieve_public_key()
            variableProcedure(tempVarLabel, indexRangeCap, _tempVar=retrieve_passkey())
            combiMatLoc = sum([_cached_caller_globals[tempVarLabel][j] * mListLoc[j] for j in range(indexRangeCap)],sp.zeros(shapeLoc, shapeLoc))
            def pairValue(j, k):
                """
                Compute the commutator [m_j, m_k] and match with the combination matrix.

                Parameters
                ----------
                j : int
                    Index of the first matrix in the commutator.
                k : int
                    Index of the second matrix in the commutator.

                Returns
                -------
                list
                    The coefficients representing the structure constants.
                """
                mat = (mListLoc[j] * mListLoc[k] - mListLoc[k] * mListLoc[j] - combiMatLoc)
                bracketVals = list(set([*mat]))
                if len(bracketVals)==1 and bracketVals[0]==0:
                    return [0]*indexRangeCap
                solLoc = list(solve_dgcv(bracketVals, _cached_caller_globals[tempVarLabel]))

                if len(solLoc) == 1:
                    result = [var.subs(solLoc[0]) for var in _cached_caller_globals[tempVarLabel]]
                    return result
                else:
                    raise Exception(
                        f"Unable to determine if matrices are closed under commutators. "
                        f"Problem matrices are in positions {j} and {k}."
                    )


            structure_data = [[[0]*indexRangeCap if k<=j else pairValue(k, j) for j in range(indexRangeCap)] for k in range(indexRangeCap)]
            for k in range(indexRangeCap):
                for j in range(k+1,indexRangeCap):
                    structure_data[k][j]=[-entry for entry in structure_data[j][k]]

            # Clear all temporary variables
            clearVar(*listVar(temporary_only=True), report=False)

            return structure_data
        else:
            raise Exception("algebraDataFromMatRep expects a list of square matrices of the same size.")
    else:
        raise Exception("algebraDataFromMatRep expects a list of square matrices.")





from ._config import get_dgcv_settings_registry
from ._dgcv_display import dgcv_init_printing


def set_dgcv_settings(theme=None,
                      format_displays=None,
                      use_latex=None,
                      version_specific_defaults=None,
                      ask_before_overwriting_objects_in_vmf=None,
                      forgo_warnings=None):
    dgcvSR = get_dgcv_settings_registry()
    if theme is not None:
        dgcvSR['theme'] = theme
    if format_displays is True:
        dgcv_init_printing()
    if use_latex is not None:
        dgcvSR['use_latex'] = use_latex
    if version_specific_defaults is not None:
        dgcvSR['version_specific_defaults'] = version_specific_defaults
    if ask_before_overwriting_objects_in_vmf is not None:
        dgcvSR['ask_before_overwriting_objects_in_vmf'] = ask_before_overwriting_objects_in_vmf
    if forgo_warnings is not None:
        dgcvSR['forgo_warnings'] = forgo_warnings


